import unittest
import tempfile

import numpy as np
import h5py

from srlife import receiver

outer_radius = 1.0
thickness = 0.1
height = 100.0

nr = 10
nt = 10
nz = 10

inner_radius = outer_radius - thickness

period = 10.0
ndays = 10

nsteps = 10

times = np.linspace(0, period * ndays, ndays * nsteps)

panel_k = 0
tube_k = np.inf


def get_temp_hdf():
    return h5py.File(tempfile.mktemp(), "w")


def valid_pressure_bc():
    return receiver.PressureBC(times, np.zeros((len(times),)))


def valid_convective_bc(loc):
    if loc == "outer":
        r = outer_radius
    else:
        r = inner_radius
    return receiver.ConvectiveBC(r, height, nz, times, np.zeros((len(times), nz)))


def valid_heatflux_bc(loc):
    if loc == "outer":
        r = outer_radius
    else:
        r = inner_radius
    return receiver.HeatFluxBC(r, height, nt, nz, times, np.zeros((len(times), nt, nz)))


def valid_fixedtemp_bc(loc):
    if loc == "outer":
        r = outer_radius
    else:
        r = inner_radius
    return receiver.FixedTempBC(
        r, height, nt, nz, times, np.zeros((len(times), nt, nz))
    )


def valid_tube(results=[]):
    tube = receiver.Tube(outer_radius, thickness, height, nr, nt, nz)

    tube.set_times(times)

    for res in results:
        tube.add_results(res, np.zeros((len(times), nr, nt, nz)))

    tube.set_bc(valid_heatflux_bc("inner"), "inner")
    tube.set_bc(valid_convective_bc("outer"), "outer")
    tube.set_pressure_bc(valid_pressure_bc())

    return tube


def valid_tube_1D(results=[]):
    tube = receiver.Tube(outer_radius, thickness, height, nr, nt, nz)

    tube.set_times(times)

    tube.make_1D(height / 2, np.pi / 3)

    for res in results:
        tube.add_results(res, np.zeros((len(times), nr)))

    return tube


def valid_panel(n=0, results=[]):
    panel = receiver.Panel(tube_k)

    for i in range(n):
        panel.add_tube(valid_tube(results))

    return panel


def valid_receiver(npanel=0, ntube=0, results=[]):
    rec = receiver.Receiver(period, ndays, panel_k)

    for i in range(npanel):
        rec.add_panel(valid_panel(ntube, results))

    return rec


class TestReceiver(unittest.TestCase):
    def test_construct(self):
        rec = valid_receiver(10, 10, ["stress", "strain"])

    def test_next_number(self):
        rec = valid_receiver(2)
        rec.add_panel(valid_panel(), name="A")
        rec.add_panel(valid_panel())

        self.assertEqual(rec.npanels, 4)

        self.assertFalse("3" in rec.panels.keys())
        self.assertTrue("2" in rec.panels.keys())

    def test_store(self):
        base = valid_receiver(10, 20, ["stress", "strain"])
        f = get_temp_hdf()
        base.save(f)
        new = receiver.Receiver.load(f)

        self.assertTrue(base.close(new))


class TestPanel(unittest.TestCase):
    def test_construct(self):
        panel = valid_panel(2)

    def test_next_number(self):
        panel = valid_panel(2)
        panel.add_tube(valid_tube(), name="A")
        panel.add_tube(valid_tube())

        self.assertEqual(panel.ntubes, 4)

        self.assertFalse("3" in panel.tubes.keys())
        self.assertTrue("2" in panel.tubes.keys())

    def test_store(self):
        base = valid_panel(5, ["stress"])

        f = get_temp_hdf()
        base.save(f)
        new = receiver.Panel.load(f)

        self.assertTrue(base.close(new))


class TestTube(unittest.TestCase):
    def test_construct(self):
        tube = valid_tube(["stress", "strain"])

    def test_invalid_size(self):
        tube = valid_tube(["stress"])
        with self.assertRaises(ValueError):
            tube.add_results("strain", np.zeros((1,)))

    def test_invalid_times(self):
        tube = valid_tube(["stress"])
        with self.assertRaises(ValueError):
            tube.set_times(np.linspace(0, 1, 10))

    def test_valid_bc(self):
        tube = valid_tube()
        tube.set_bc(valid_heatflux_bc("outer"), "outer")
        tube.set_bc(valid_heatflux_bc("inner"), "inner")

    def test_invalid_bc(self):
        tube = valid_tube()
        with self.assertRaises(ValueError):
            tube.set_bc(valid_heatflux_bc("outer"), "inner")

    def test_invalid_bc_location(self):
        tube = valid_tube()
        with self.assertRaises(ValueError):
            tube.set_bc(valid_heatflux_bc("outer"), "blah")

    def test_store(self):
        tube = valid_tube(["stress"])
        tube.set_bc(valid_heatflux_bc("outer"), "outer")

        f = get_temp_hdf()
        tube.save(f)
        new = receiver.Tube.load(f)

        self.assertTrue(tube.close(new))

    def test_store_1d(self):
        tube = valid_tube_1D(["stress"])
        tube.set_bc(valid_heatflux_bc("outer"), "outer")

        f = get_temp_hdf()
        tube.save(f)
        new = receiver.Tube.load(f)

        self.assertTrue(tube.close(new))


class TestPressureBC(unittest.TestCase):
    """
    Basic setup for PressureBC
    """

    def test_construct(self):
        bc = valid_pressure_bc()

    def test_store(self):
        f = get_temp_hdf()
        orig = valid_pressure_bc()
        orig.save(f)
        new = receiver.PressureBC.load(f)
        self.assertTrue(orig.close(new))


class TestHeatFluxBC(unittest.TestCase):
    """
    Test basic setup of a HeatFluxBC
    """

    def test_construct(self):
        bc = valid_heatflux_bc("outer")

    def test_wrong_size(self):
        with self.assertRaises(ValueError):
            bc = receiver.HeatFluxBC(
                inner_radius, height, nt, nz, times, np.zeros((1,))
            )

    def test_store(self):
        f = get_temp_hdf()
        orig = valid_heatflux_bc("outer")
        orig.save(f)
        new = receiver.ThermalBC.load(f)
        self.assertTrue(orig.close(new))


class TestFixedTempBC(unittest.TestCase):
    """
    Test basic setup of a FixedTempBC
    """

    def test_construct(self):
        bc = valid_fixedtemp_bc("outer")

    def test_wrong_size(self):
        with self.assertRaises(ValueError):
            bc = receiver.FixedTempBC(
                inner_radius, height, nt, nz, times, np.zeros((1,))
            )

    def test_store(self):
        f = get_temp_hdf()
        orig = valid_fixedtemp_bc("outer")
        orig.save(f)
        new = receiver.ThermalBC.load(f)
        self.assertTrue(orig.close(new))


class TestConvectiveBC(unittest.TestCase):
    """
    Test basic setup of a ConvectiveBC
    """

    def test_construct(self):
        bc = valid_convective_bc("outer")

    def test_wrong_size(self):
        with self.assertRaises(ValueError):
            bc = receiver.ConvectiveBC(inner_radius, height, nz, times, np.zeros((1,)))

    def test_store(self):
        f = get_temp_hdf()
        orig = valid_convective_bc("outer")
        orig.save(f)
        new = receiver.ThermalBC.load(f)
        self.assertTrue(orig.close(new))



import sys

sys.path.append("../../..")

import numpy as np
import matplotlib.pyplot as plt

from srlife import thermal, receiver, materials


class ManufacturedSolution:
    """
    A manufactured heat transport solution
    """

    def __init__(self, name, dim, solution, source, steady=False):
        """
        Parameters:
          name:       descriptive name
          dim:        dimension of solution
          solution:   function of t, r, ...
          source      function of t, k, alpha, r, ...
          flux        function of t, k, alpha, r, ... (for BC)
        """
        self.name = name
        self.dim = dim
        self.soln = solution
        self.source = source
        self.steady = steady

    def solve(
        self,
        solver,
        thermal,
        fluid,
        r=1.0,
        t=0.2,
        h=1,
        time=1,
        ntime=11,
        nr=11,
        nt=20,
        nz=10,
        T0=0.0,
    ):
        """
        Generate the appropriate tube and solve with the provided solver

        Parameters:
          solver:         the thermal solver to test
          thermal:        the thermal model to test

        Other Parameters:
          r:              tube radius
          t:              tube thickness
          h:              tube height
          time:           maximum time
          ntime:          number of time steps
          nr:             number of radial increments
          nt:             number of circumferential increments
          nz:             number of axial increments
          T0:             initial temperature
        """
        solver.steady = self.steady

        tube = receiver.Tube(r, t, h, nr, nt, nz, T0)

        times = np.linspace(0, time, ntime)
        tube.set_times(times)

        if self.dim == 2:
            tube.make_2D(tube.h / 2)
        elif self.dim == 1:
            tube.make_1D(tube.h / 2, 0)

        def T0(*args):
            return self.soln(0, *args)

        def sfn(t, *args):
            k = thermal.conductivity(self.soln(t, *args))
            a = thermal.diffusivity(self.soln(t, *args))

            return self.source(t, k, a, *args)

        solver.solve(
            tube,
            thermal,
            materials.ConstantFluidMaterial({}),
            source=sfn,
            T0=T0,
            fix_edge=self.soln,
        )

        return tube

    def plot_comparison(self, soln):
        """
        Plot a comparison to a manufactured solution

        Parameters:
          soln:           particular tube object with the solution
        """
        mesh = self._generate_mesh(
            soln.r, soln.t, soln.h, soln.times, soln.nr, soln.nt, soln.nz
        )
        T = self.soln(*mesh)

        print(self.name)
        print(
            "Max absolute error: %e" % np.max(np.abs(T - soln.results["temperature"]))
        )
        keep = np.abs(T) > 0
        print(
            "Max relative error: %e"
            % np.max(
                np.abs(T[keep] - soln.results["temperature"][keep]) / np.abs(T[keep])
            )
        )
        print("")

        plt.figure()
        if self.dim == 1:
            plot = [1, soln.nr // 2, -2]
            for p in plot:
                (l,) = plt.plot(mesh[0][:, p], soln.results["temperature"][:, p])
                plt.plot(mesh[0][:, p], T[:, p], ls="--", color=l.get_color())
            plt.xlabel("Time")
            plt.ylabel("Temperature")
            plt.title("1D: left, middle, right")
        elif self.dim == 2:
            plot_r = [0, 1, soln.nr // 2, -2, -1]
            plot_t = [0, 1, soln.nt // 2, -2, -1]

            for rp in plot_r:
                for tp in plot_t:
                    (l,) = plt.plot(
                        mesh[0][:, rp, tp], soln.results["temperature"][:, rp, tp]
                    )
                    plt.plot(
                        mesh[0][:, rp, tp], T[:, rp, tp], ls="--", color=l.get_color()
                    )
            plt.xlabel("Time")
            plt.ylabel("Temperature")
            plt.title("2D, r/theta slices")
        elif self.dim == 3:
            plot_r = [0, 1, soln.nr // 2, -2, -1]
            plot_t = [0, 1, soln.nt // 2, -2, -1]
            plot_z = [0, 1, soln.nz // 2, -2, -1]
            for rp in plot_r:
                for tp in plot_t:
                    for zp in plot_z:
                        (l,) = plt.plot(
                            mesh[0][:, rp, tp, zp],
                            soln.results["temperature"][:, rp, tp, zp],
                        )
                        plt.plot(
                            mesh[0][:, rp, tp, zp],
                            T[:, rp, tp, zp],
                            ls="--",
                            color=l.get_color(),
                        )
            plt.xlabel("Time")
            plt.ylabel("Temperature")
            plt.title("3D, r/theta/z slices")

    def assess_comparison(self, soln, tol, atol):
        """
        Return true if solution matches, false otherwise

        Parameters:
          soln:           particular tube object with the solution
          tol:            relative tolerance
        """
        mesh = self._generate_mesh(
            soln.r, soln.t, soln.h, soln.times, soln.nr, soln.nt, soln.nz
        )
        T = self.soln(*mesh)

        return np.allclose(T, soln.results["temperature"], rtol=tol, atol=atol)

    def _generate_mesh(self, r, t, h, times, nr, nt, nz):
        """
        Generate the appropriate finite difference mesh for a particular problem

        Parameters:
          r:           radius
          t:           thickness
          h            height
          times:       discrete time steps
          nr:          number of radial increments
          nt:          number of circumferential increments
          nz:          number of axial increments
        """
        rs = np.linspace(r - t, r, nr)
        ts = np.linspace(0, 2 * np.pi, nt + 1)[:-1]
        zs = np.linspace(0, h, nz)

        geom = [rs, ts, zs]

        return np.meshgrid(times, *geom[: self.dim], indexing="ij")

    def _generate_surface_mesh(self, t, h, times, nt, nz):
        """
        Generate the appropriate finite difference mesh for a particular problem

        Parameters:
          t:           thickness
          h:           height
          times:       discrete time steps
          nr:          number of radial increments
          nt:          number of circumferential increments
          nz:          number of axial increments
        """
        ts = np.linspace(0, 2 * np.pi, nt + 1)[:-1]
        zs = np.linspace(0, h, nz)

        geom = [ts, zs]

        return np.meshgrid(times, *geom, indexing="ij")



# pylint: disable=no-member
"""
  This module contains methods for calculating the reliability and
  creep-fatigue damage given completely-solved tube results and
  damage material properties
"""

import numpy as np
import numpy.linalg as la
import scipy.optimize as opt
import multiprocess


class WeibullFailureModel:
    """Parent class for time independent Weibull failure models

    Determines principal stresses from mandel stress

    Determines tube reliability and overall reliability by taking input of
    element log reliabilities from respective Weibull failure model
    """

    def __init__(self, pset, *args, cares_cutoff=True):
        """Initialize the Weibull Failure Model

        Boolean:
        cares_cutoff:    condition for forcing reliability as unity in case of
                        high compressive stresses
        """
        self.cares_cutoff = cares_cutoff

    def calculate_principal_stress(self, stress):
        """
        Calculate the principal stresses from Mandel vector and converts
        to conventional notation
        """
        if stress.ndim == 2:
            tensor = np.zeros(
                stress.shape[:1] + (3, 3)
            )  # [:1] when no time steps involved
        elif stress.ndim == 3:
            tensor = np.zeros(
                stress.shape[:2] + (3, 3)
            )  # [:2] when time steps involved
        # indices where (0,0) => (1,1)
        inds = [
            [(0, 0)],
            [(1, 1)],
            [(2, 2)],
            [(1, 2), (2, 1)],
            [(0, 2), (2, 0)],
            [(0, 1), (1, 0)],
        ]
        # multiplicative factors
        mults = [1.0, 1.0, 1.0, np.sqrt(2), np.sqrt(2), np.sqrt(2)]

        # Converting mandel notation to conventional notation
        for i, (grp, m) in enumerate(zip(inds, mults)):
            for a, b in grp:
                tensor[..., a, b] = stress[..., i] / m

        return la.eigvalsh(tensor)

    def determine_reliability(
        self, receiver, material, nthreads=1, decorator=lambda x, n: x
    ):
        """
        Determine the reliability of the tubes in the receiver by calculating individual
        material point reliabilities and finding the minimum of all points.

        Parameters:
          receiver        fully-solved receiver object
          material        material model to use

        Additional Parameters:
          nthreads        number of threads
          decorator       progress bar
        """
        with multiprocess.Pool(nthreads) as p:
            results = list(
                decorator(
                    p.imap(
                        lambda x: self.tube_log_reliability(x, material, receiver),
                        receiver.tubes,
                    ),
                    receiver.ntubes,
                )
            )

        p_tube = np.array([res[0] for res in results])
        tube_fields = [res[1] for res in results]

        # Tube reliability is the minimum of all the time steps
        tube = np.min(p_tube, axis=1)

        # Overall reliability is the minimum of the sum
        overall = np.min(np.sum(p_tube, axis=0))

        # Add the field to the tubes
        for tubei, field in zip(receiver.tubes, tube_fields):
            tubei.add_quadrature_results("log_reliability", field)

        # Convert back from log-prob as we go
        return {
            "tube_reliability": np.exp(tube),
            "overall_reliability": np.exp(overall),
        }

    def tube_log_reliability(self, tube, material, receiver):
        """
        Calculate the log reliability of a single tube
        """
        volumes = tube.element_volumes()

        stresses = np.transpose(
            np.mean(
                np.stack(
                    (
                        tube.quadrature_results["stress_xx"],
                        tube.quadrature_results["stress_yy"],
                        tube.quadrature_results["stress_zz"],
                        tube.quadrature_results["stress_yz"],
                        tube.quadrature_results["stress_xz"],
                        tube.quadrature_results["stress_xy"],
                    )
                ),
                axis=-1,
            ),
            axes=(1, 2, 0),
        )

        temperatures = np.mean(tube.quadrature_results["temperature"], axis=-1)

        # Do it this way so we can vectorize
        inc_prob = self.calculate_element_log_reliability(
            stresses, temperatures, volumes, material
        )

        # CARES/LIFE cutoff
        if self.cares_cutoff:
            pstress = self.calculate_principal_stress(stresses).reshape(-1, 3)
            pmax = np.max(pstress, axis=1)
            pmin = np.min(pstress, axis=1)
            remove = np.abs(pmin / (pmax + 1.0e-16)) > 3.0
            mod_prob = inc_prob.flatten()
            mod_prob[remove] = 0.0
            inc_prob = mod_prob.reshape(inc_prob.shape)

        # Return the sums as a function of time along with the field itself
        return np.sum(inc_prob, axis=1), np.transpose(
            np.stack((inc_prob, inc_prob)), axes=(1, 2, 0)
        )


class CrackShapeIndependent(WeibullFailureModel):
    """
    Parent class for crack shape independent models
    which include only PIA and WNTSA models

    Determines normal stress acting on crack
    """

    def __init__(self, pset, *args, **kwargs):
        """
        Create a mesh grid of angles that can represent crack orientations
        Evaluate direction cosines using the angles

        Default values given to nalpha and nbeta which are the number of
        segments in the mesh grid
        """
        super().__init__(pset, *args, **kwargs)

        # limits and number of segments for angles
        self.nalpha = pset.get_default("nalpha", 21)
        self.nbeta = pset.get_default("nbeta", 31)

        # Mesh grid of the vectorized angle values
        self.A, self.B = np.meshgrid(
            np.linspace(0, np.pi, self.nalpha),
            np.linspace(0, 2 * np.pi, self.nbeta, endpoint=False),
            indexing="ij",
        )

        # Increment of angles to be used in evaluating integral
        self.dalpha = (self.A[-1, -1] - self.A[0, 0]) / (self.nalpha - 1)
        self.dbeta = (self.B[-1, -1] - self.B[0, 0]) / (self.nbeta - 1)

        # Direction cosines
        self.l = np.cos(self.A)
        self.m = np.sin(self.A) * np.cos(self.B)
        self.n = np.sin(self.A) * np.sin(self.B)

    def calculate_normal_stress(self, mandel_stress):
        """
        Use direction cosines to calculate the normal stress to
        a crack given the Mandel vector
        """
        # Principal stresses
        pstress = self.calculate_principal_stress(mandel_stress)

        # Normal stress
        return (
            pstress[..., 0, None, None] * (self.l**2)
            + pstress[..., 1, None, None] * (self.m**2)
            + pstress[..., 2, None, None] * (self.n**2)
        )


class CrackShapeDependent(WeibullFailureModel):
    """
    Parent class for crack shape dependent models

    Determines normal, shear, total and equivanlent stresses acting on cracks

    Calculates the element reliability using the equivalent stress from the
    crack shape dependent models
    """

    def __init__(self, pset, *args, **kwargs):
        """
        Create a mesh grid of angles that can represent crack orientations
        Evaluate direction cosines using the angles

        Default values given to nalpha and nbeta which are the number of
        segments in the mesh grid
        """
        super().__init__(pset, *args, **kwargs)

        # limits and number of segments for angles
        self.nalpha = pset.get_default("nalpha", 121)
        self.nbeta = pset.get_default("nbeta", 121)

        # Mesh grid of the vectorized angle values
        self.A, self.B = np.meshgrid(
            np.linspace(0, np.pi / 2, self.nalpha),
            np.linspace(0, np.pi / 2, self.nbeta),
            indexing="ij",
        )

        # Increment of angles to be used in evaluating integral
        self.dalpha = (self.A[-1, -1] - self.A[0, 0]) / (self.nalpha - 1)
        self.dbeta = (self.B[-1, -1] - self.B[0, 0]) / (self.nbeta - 1)

        # Direction cosines
        self.l = np.cos(self.A)
        self.m = np.sin(self.A) * np.cos(self.B)
        self.n = np.sin(self.A) * np.sin(self.B)

    def calculate_normal_stress(self, mandel_stress):
        """
        Use direction cosines to calculate the normal stress to
        a crack given the Mandel vector
        """
        # Principal stresses
        pstress = self.calculate_principal_stress(mandel_stress)

        # Normal stress
        return (
            pstress[..., 0, None, None] * (self.l**2)
            + pstress[..., 1, None, None] * (self.m**2)
            + pstress[..., 2, None, None] * (self.n**2)
        )

    def calculate_total_stress(self, mandel_stress):
        """
        Calculate the total stress given the Mandel vector
        """
        # Principal stresses
        pstress = self.calculate_principal_stress(mandel_stress)

        # Total stress
        return np.sqrt(
            ((pstress[..., 0, None, None] * self.l) ** 2)
            + ((pstress[..., 1, None, None] * self.m) ** 2)
            + ((pstress[..., 2, None, None] * self.n) ** 2)
        )

    def calculate_shear_stress(self, mandel_stress):
        """
        Calculate the shear stress given the normal and total stress
        """
        # Normal stress
        sigma_n = self.calculate_normal_stress(mandel_stress)

        # Total stress
        sigma = self.calculate_total_stress(mandel_stress)

        # Shear stress
        with np.errstate(invalid="ignore"):
            return np.sqrt(sigma**2 - sigma_n**2)

    def calculate_flattened_eq_stress(
        self, mandel_stress, temperatures, material, A, dalpha, dbeta
    ):
        """
        Calculate the integral of equivalent stresses given the material
        properties and integration limits

        Parameters:
          mandel_stress:  element stresses in Mandel convention
          temperatures:   element temperatures
          material        material model to use

        Additional Parameters:
          A:              mesh grid of vectorized angle values
          dalpha:         increment of angle alpha to be used in evaluating integral
          dbeta:          increment of angle beta to be used in evaluating integral
        """
        self.temperatures = temperatures
        self.material = material
        self.mandel_stress = mandel_stress
        self.mvals = self.material.modulus(temperatures)

        # Principal stresses
        pstress = self.calculate_principal_stress(mandel_stress)

        # Projected equivalent stresses
        sigma_e = self.calculate_eq_stress(
            mandel_stress, self.temperatures, self.material
        )

        # Suppressing warning given when negative numbers are raised to rational numbers
        with np.errstate(invalid="ignore"):
            # Defining area integral element wise
            integral = (
                (sigma_e ** self.mvals[..., None, None])
                * np.sin(self.A)
                * self.dalpha
                * self.dbeta
            )

        # Flatten the last axis and calculate the mean of the positive values along that axis
        if pstress.ndim == 2:
            flat = integral.reshape(
                integral.shape[:1] + (-1,)
            )  # when no time steps involved
        elif pstress.ndim == 3:
            flat = integral.reshape(
                integral.shape[:2] + (-1,)
            )  # when time steps involved

        # Suppressing warning to ignoring initial NaN values
        with np.errstate(invalid="ignore"):
            # Summing over area integral elements ignoring NaN values
            flat = np.nansum(flat, axis=-1)

        return flat

    def calculate_element_log_reliability(
        self, mandel_stress, temperatures, volumes, material
    ):
        """
        Calculate the element log reliability given the equivalent stress

        Parameters:
          mandel_stress:  element stresses in Mandel convention
          temperatures:   element temperatures
          volumes:        element volumes
          material:       material model  object with required data
        """
        self.temperatures = temperatures
        self.material = material
        self.mandel_stress = mandel_stress

        # Material parameters
        self.svals = self.material.strength(temperatures)
        self.mvals = self.material.modulus(temperatures)
        self.kvals = self.svals ** (-self.mvals)
        self.kpvals = (2 * self.mvals + 1) * self.kvals

        # Equivalent stress raied to exponent mv
        flat = (
            self.calculate_flattened_eq_stress(
                mandel_stress,
                self.temperatures,
                self.material,
                self.A,
                self.dalpha,
                self.dbeta,
            )
        ) ** (1 / self.mvals)

        return -(2 * self.kpvals / np.pi) * (flat**self.mvals) * volumes


class PIAModel(CrackShapeIndependent):
    """
    Principal of independent action failure model

    Calculates reliability using only tensile stresses
    that are assumed to act independently on cracks
    """

    def calculate_element_log_reliability(
        self, mandel_stress, temperatures, volumes, material
    ):
        """
        Calculate the element log reliability

        Parameters:
          mandel_stress:  element stresses in Mandel convention
          temperatures:   element temperatures
          volumes:        element volumes
          material:       material model object with required data that includes
                          Weibull scale parameter (svals) and Weibull modulus (mvals)
        """
        # Material parameters
        svals = material.strength(temperatures)
        mvals = material.modulus(temperatures)
        kvals = svals ** (-mvals)

        # Principal stresses
        pstress = self.calculate_principal_stress(mandel_stress)

        # Only tension
        pstress[pstress < 0] = 0

        return -kvals * np.sum(pstress ** mvals[..., None], axis=-1) * volumes


class WNTSAModel(CrackShapeIndependent):
    """
    Weibull normal tensile average failure model

    Evaluates an average normal tensile stress (acting on the cracks) integrated over the
    area of a sphere, and uses it to calculate the element reliability
    """

    def calculate_avg_normal_stress(self, mandel_stress, temperatures, material):
        """
        Calculate the average normal tensile stresses from the pricipal stresses

        Parameters:
          mandel_stress:  element stresses in Mandel convention
          temperatures:   element temperatures
          material:       material model object with required data
        """
        # Material parameters
        mvals = material.modulus(temperatures)

        # Principal stresses
        pstress = self.calculate_principal_stress(mandel_stress)

        # Normal stress
        sigma_n = self.calculate_normal_stress(mandel_stress)

        # Area integral; suppressing warning given when negative numbers are raised
        # to rational numbers
        with np.errstate(invalid="ignore"):
            integral = (
                (sigma_n ** mvals[..., None, None])
                * np.sin(self.A)
                * self.dalpha
                * self.dbeta
            ) / (4 * np.pi)

        # Flatten the last axis and calculate the mean of the positive values along that axis
        if pstress.ndim == 2:
            flat = integral.reshape(
                integral.shape[:1] + (-1,)
            )  # when no time steps involved
        elif pstress.ndim == 3:
            flat = integral.reshape(
                integral.shape[:2] + (-1,)
            )  # when time steps involved

        # Suppressing warning to ignoring initial NaN values
        with np.errstate(invalid="ignore"):
            # Average stress
            return np.nansum(np.where(flat >= 0.0, flat, np.nan), axis=-1)

    def calculate_element_log_reliability(
        self, mandel_stress, temperatures, volumes, material
    ):
        """
        Calculate the element log reliability

        Parameters:
          mandel_stress:  element stresses in Mandel convention
          temperatures:   element temperatures
          volumes:        element volumes
          material:       material model  object with required data
        """

        # Material parameters
        svals = material.strength(temperatures)
        mvals = material.modulus(temperatures)
        kvals = svals ** (-mvals)
        kpvals = (2 * mvals + 1) * kvals

        # Average normal tensile stress raied to exponent mv
        avg_nstress = (
            self.calculate_avg_normal_stress(mandel_stress, temperatures, material)
        ) ** (1 / mvals)

        return -kpvals * (avg_nstress**mvals) * volumes


class MTSModelGriffithFlaw(CrackShapeDependent):
    """
    Maximum tensile stess failure model with a Griffith flaw

    Evaluates an equivalent stress (acting on the cracks) using the normal and
    shear stresses in the maximum tensile stress fracture criterion for a Griffith flaw
    """

    def calculate_eq_stress(self, mandel_stress, temperatures, material):
        """
        Calculate the equivalent stresses from the normal and shear stresses
        """
        # Normal stress
        sigma_n = self.calculate_normal_stress(mandel_stress)

        # Shear stress
        tau = self.calculate_shear_stress(mandel_stress)

        # Projected equivalent stress
        return 0.5 * (sigma_n + np.sqrt((sigma_n**2) + (tau**2)))


class MTSModelPennyShapedFlaw(CrackShapeDependent):
    """
    Maximum tensile stess failure model with a penny shaped flaw

    Evaluates an equivalent stress (acting on the cracks) using the normal and
    shear stresses in the maximum tensile stress fracture criterion for a penny shaped flaw
    """

    def calculate_eq_stress(self, mandel_stress, temperatures, material):
        """
        Calculate the average normal tensile stresses from the normal and shear stresses

        Additional parameter:
        nu:     Poisson ratio
        """

        # Material parameters
        nu = material.nu(temperatures).flat[0]

        sigma_n = self.calculate_normal_stress(mandel_stress)
        tau = self.calculate_shear_stress(mandel_stress)
        # Projected equivalent stress
        return 0.5 * (
            sigma_n + np.sqrt((sigma_n**2) + ((tau / (1 - (0.5 * nu))) ** 2))
        )


class CSEModelGriffithFlaw(CrackShapeDependent):
    """
    Coplanar strain energy failure model with a Griffith flaw

    Evaluates an equivalent stress (acting on the cracks) using the normal and
    shear stresses in the coplanar strain energy fracture criterion
    for a Griffith flaw
    """

    def calculate_eq_stress(self, mandel_stress, temperatures, material):
        """
        Calculate the equivalent stresses from the normal and shear stresses
        """
        # Normal stress
        sigma_n = self.calculate_normal_stress(mandel_stress)

        # Shear stress
        tau = self.calculate_shear_stress(mandel_stress)

        # Projected equivalent stress
        return np.sqrt((sigma_n**2) + (tau**2))


class CSEModelPennyShapedFlaw(CrackShapeDependent):
    """
    Coplanar strain energy failure model with a penny shaped flaw

    Evaluates an equivalent stress (acting on the cracks) using the normal and
    shear stresses in the coplanar strain energy fracture criterion
    for a penny shaped flaw
    """

    def calculate_eq_stress(self, mandel_stress, temperatures, material):
        """
        Calculate the equivalent stresses from the normal and shear stresses

        Additional parameter:
        nu:     Poisson ratio
        """

        # Material parameters
        nu = material.nu(temperatures).flat[0]

        # Normal stress
        sigma_n = self.calculate_normal_stress(mandel_stress)

        # Shear stress
        tau = self.calculate_shear_stress(mandel_stress)

        # Projected equivalent stress
        return np.sqrt((sigma_n**2) + (tau / (1 - (0.5 * nu))) ** 2)


class SMMModelGriffithFlaw(CrackShapeDependent):
    """
    Shetty mixed-mod failure model with a Griffith flaw

    Evaluates an equivalent stress (acting on the cracks) using the normal and
    shear stresses in the Shetty mixed-mode fracture criterion
    for a Griffith flaw
    """

    def calculate_eq_stress(self, mandel_stress, temperatures, material):
        """
        Calculate the equivalent stresses from the normal and shear stresses

        Additional parameters:
        cbar:   Emperical constant
        """

        # Material parameters
        cbar = material.c_bar(temperatures).flat[0]

        # Normal stress
        sigma_n = self.calculate_normal_stress(mandel_stress)

        # Shear stress
        tau = self.calculate_shear_stress(mandel_stress)

        # Projected equivalent stress
        return 0.5 * (sigma_n + np.sqrt((sigma_n**2) + ((2 * tau / cbar) ** 2)))


class SMMModelPennyShapedFlaw(CrackShapeDependent):
    """
    Shetty mixed mode failure model with a penny shaped flaw

    Evaluates an equivalent stress (acting on the cracks) using the normal and
    shear stresses in the Shetty mixed-mode fracture criterion
    for a Griffith flaw
    """

    def calculate_eq_stress(self, mandel_stress, temperatures, material):
        """
        Calculate the equivalent stresses from the normal and shear stresses

        Additional parameters:
        nu:     Poisson ratio
        cbar:   Emperical constant
        """

        # Material parameters
        nu = material.nu(temperatures).flat[0]
        cbar = material.c_bar(temperatures).flat[0]

        # Normal stress
        sigma_n = self.calculate_normal_stress(mandel_stress)

        # Shear stress
        tau = self.calculate_shear_stress(mandel_stress)

        # Projected equivalent stress
        return 0.5 * (
            sigma_n + np.sqrt((sigma_n**2) + ((4 * tau / (cbar * (2 - nu))) ** 2))
        )


class DamageCalculator:
    """
    Parent class for all damage calculators, handling common iteration
    and scaling options
    """

    def __init__(self, pset):
        """
        Parameters:
          pset:       damage parameters
        """
        self.extrapolate = pset.get_default("extrapolate", "lump")
        self.order = pset.get_default("order", 1)

    def single_cycles(self, tube, material, receiver):
        """
        Calculate damage for a single tube

        Parameters:
          tube:       fully-populated tube object
          material:   damage material
          receiver:   receiver object (for metadata)
        """
        raise NotImplementedError("Superclass not implemented")

    def determine_life(self, receiver, material, nthreads=1, decorator=lambda x, n: x):
        """
        Determine the life of the receiver by calculating individual
        material point damage and finding the minimum of all points.

        Parameters:
          receiver        fully-solved receiver object
          material        material model ot use

        Additional Parameters:
          nthreads        number of threads
          decorator       progress bar
        """
        # pylint: disable=no-member
        with multiprocess.Pool(nthreads) as p:
            Ns = list(
                decorator(
                    p.imap(
                        lambda x: self.single_cycles(x, material, receiver),
                        receiver.tubes,
                    ),
                    receiver.ntubes,
                )
            )
        N = min(Ns)

        # Results come out as days
        return N

    def make_extrapolate(self, D):
        """
        Return a damage extrapolation function based on self.extrapolate
        giving the damage for the nth cycle

        Parameters:
          D:      raw, per cycle damage
        """
        if self.extrapolate == "lump":
            return lambda N, D=D: N * np.sum(D) / len(D)
        elif self.extrapolate == "last":

            def Dfn(N, D=D):
                N = int(N)
                if N < len(D) - 1:
                    return np.sum(D[:N])
                else:
                    return np.sum(D[:-1]) + D[-1] * N

            return Dfn
        elif self.extrapolate == "poly":
            p = np.polyfit(np.array(list(range(len(D)))) + 1, D, self.order)
            return lambda N, p=p: np.polyval(p, N)
        else:
            raise ValueError(
                "Unknown damage extrapolation approach %s!" % self.extrapolate
            )


class TimeFractionInteractionDamage(DamageCalculator):
    """
    Calculate life using the ASME time-fraction type approach
    """

    def single_cycles(self, tube, material, receiver):
        """
        Calculate the single-tube number of repetitions to failure

        Parameters:
          tube        single tube with full results
          material    damage material model
          receiver    receiver, for metadata
        """
        # Material point cycle creep damage
        Dc = self.creep_damage(tube, material, receiver)

        # Material point cycle fatigue damage
        Df = self.fatigue_damage(tube, material, receiver)

        nc = receiver.days

        # This is going to be expensive, but I don't see much way around it
        return min(
            self.calculate_max_cycles(
                self.make_extrapolate(c), self.make_extrapolate(f), material
            )
            for c, f in zip(Dc.reshape(nc, -1).T, Df.reshape(nc, -1).T)
        )

    def calculate_max_cycles(self, Dc, Df, material, rep_min=1, rep_max=1e6):
        """
        Actually calculate the maximum number of repetitions for a single point

        Parameters:
          Dc          creep damage per simulated cycle
          Df          fatigue damage per simulated cycle
          material    damaged material properties
        """
        if not material.inside_envelope("cfinteraction", Df(rep_min), Dc(rep_min)):
            return 0

        if material.inside_envelope("cfinteraction", Df(rep_max), Dc(rep_max)):
            return np.inf

        return opt.brentq(
            lambda N: material.inside_envelope("cfinteraction", Df(N), Dc(N)) - 0.5,
            rep_min,
            rep_max,
        )

    def creep_damage(self, tube, material, receiver):
        """
        Calculate creep damage at each material point

        Parameters:
          tube        single tube with full results
          material    damage material model
          receiver    receiver, for metadata
        """
        # For now just use the von Mises effective stress
        vm = np.sqrt(
            (
                (
                    tube.quadrature_results["stress_xx"]
                    - tube.quadrature_results["stress_yy"]
                )
                ** 2.0
                + (
                    tube.quadrature_results["stress_yy"]
                    - tube.quadrature_results["stress_zz"]
                )
                ** 2.0
                + (
                    tube.quadrature_results["stress_zz"]
                    - tube.quadrature_results["stress_xx"]
                )
                ** 2.0
                + 6.0
                * (
                    tube.quadrature_results["stress_xy"] ** 2.0
                    + tube.quadrature_results["stress_yz"] ** 2.0
                    + tube.quadrature_results["stress_xz"] ** 2.0
                )
            )
            / 2.0
        )

        tR = material.time_to_rupture(
            "averageRupture", tube.quadrature_results["temperature"], vm
        )
        dts = np.diff(tube.times)
        time_dmg = dts[:, np.newaxis, np.newaxis] / tR[1:]

        # Break out to cycle damage
        inds = self.id_cycles(tube, receiver)

        cycle_dmg = np.array(
            [
                np.sum(time_dmg[inds[i] : inds[i + 1]], axis=0)
                for i in range(receiver.days)
            ]
        )

        return cycle_dmg

    def fatigue_damage(self, tube, material, receiver):
        """
        Calculate fatigue damage at each material point

        Parameters:
          tube        single tube with full results
          material    damage material model
          receiver    receiver, for metadata
        """
        # Identify cycle boundaries
        inds = self.id_cycles(tube, receiver)

        # Run through each cycle and ID max strain range and fatigue damage
        strain_names = [
            "mechanical_strain_xx",
            "mechanical_strain_yy",
            "mechanical_strain_zz",
            "mechanical_strain_yz",
            "mechanical_strain_xz",
            "mechanical_strain_xy",
        ]
        strain_factors = [1.0, 1.0, 1.0, 2.0, 2.0, 2.0]

        cycle_dmg = np.array(
            [
                self.cycle_fatigue(
                    np.array(
                        [
                            ef * tube.quadrature_results[en][inds[i] : inds[i + 1]]
                            for en, ef in zip(strain_names, strain_factors)
                        ]
                    ),
                    tube.quadrature_results["temperature"][inds[i] : inds[i + 1]],
                    material,
                )
                for i in range(receiver.days)
            ]
        )

        return cycle_dmg

    def id_cycles(self, tube, receiver):
        """
        Helper to separate out individual cycles by index

        Parameters:
          tube        single tube with results
          receiver    receiver, for metadata
        """
        tm = np.mod(tube.times, receiver.period)
        inds = list(np.where(tm == 0)[0])
        if len(inds) != (receiver.days + 1):
            raise ValueError(
                "Tube times not compatible with the receiver"
                " number of days and cycle period!"
            )

        return inds

    def cycle_fatigue(self, strains, temperatures, material, nu=0.5):
        """
        Calculate fatigue damage for a single cycle

        Parameters:
          strains         single cycle strains
          temperatures    single cycle temperatures
          material        damage model

        Additional parameters:
          nu              effective Poisson's ratio to use
        """
        pt_temps = np.max(temperatures, axis=0)

        pt_eranges = np.zeros(pt_temps.shape)

        nt = strains.shape[1]
        for i in range(nt):
            for j in range(nt):
                de = strains[:, j] - strains[:, i]
                eq = (
                    np.sqrt(2)
                    / (2 * (1 + nu))
                    * np.sqrt(
                        (de[0] - de[1]) ** 2
                        + (de[1] - de[2]) ** 2
                        + (de[2] - de[0]) ** 2.0
                        + 3.0 / 2.0 * (de[3] ** 2.0 + de[4] ** 2.0 + de[5] ** 2.0)
                    )
                )
                pt_eranges = np.maximum(pt_eranges, eq)

        dmg = np.zeros(pt_eranges.shape)
        # pylint: disable=not-an-iterable
        for ind in np.ndindex(*dmg.shape):
            dmg[ind] = 1.0 / material.cycles_to_fail(
                "nominalFatigue", pt_temps[ind], pt_eranges[ind]
            )

        return dmg



import unittest

from srlife import library

fluids = ["salt"]
alloys = ["316H", "800H", "A617", "740H", "A282", "A230"]
thermalfluids = ["32MgCl2-68KCl.xml"]


class TestMaterials(unittest.TestCase):
    def test_load_materials(self):
        for alloy in alloys:
            thermal, deformation, damage = library.load_material(
                alloy, "base", "base", "base"
            )
            deformation.get_neml_model()
            thermal, deformation, damage = library.load_material(
                alloy, "base", "elastic_creep", "base"
            )
            deformation.get_neml_model()
            thermal, deformation, damage = library.load_material(
                alloy, "base", "elastic_model", "base"
            )
            deformation.get_neml_model()

        def test_load_fluids(self):
            for fluid in fluids:
                library.load_fluid(fluid, "base")



#!/usr/bin/env python3

import sys

sys.path.append("../../..")
from srlife import materials, thermal

from thermalsol import ManufacturedSolution

import numpy as np

problem = ManufacturedSolution(
    "test",
    3,
    lambda t, r, th, z: np.log(r) * np.sin(th) * np.cos(z) / (t + 1.0),
    lambda t, k, alpha, r, th, z: k
    * np.log(r)
    * np.sin(th)
    * np.cos(z)
    / (t + 1)
    * (1.0 + 1 / r**2.0 - 1.0 / (alpha * (t + 1))),
)


def run_with(solver, material, fluid):
    """
    Run the standard problems with the provided solver and material
    """
    for problem in problems:
        res = problem.solve(solver, material, fluid)
        problem.plot_comparison(res)
        plt.show()


if __name__ == "__main__":
    solver = thermal.FiniteDifferenceImplicitThermalSolver()
    tmat = materials.ConstantThermalMaterial("Test", 10.0, 5.0)
    fmat = materials.ConstantFluidMaterial({"Test": 7.5})

    if len(sys.argv) == 2:
        n = int(sys.argv[1])
    else:
        n = 1

    for i in range(n):
        res = problem.solve(solver, tmat, fmat)



#!/usr/bin/env python3

import sys

sys.path.append("../../..")

import random as ra

import matplotlib.pyplot as plt

from srlife import spring, structural, receiver

from neml import models, elasticity

import unittest

import numpy as np
import numpy.linalg as la
import networkx as nx


class Network(nx.MultiGraph):
    """
    Simplified version of panel network where everything is an elastic spring

    Boundary conditions are always fixed on the bottom and free on the top
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if len(args) == 0:
            self.add_node("top", distance=0)
            self.cpanel = 0

    def add_panel(self, k):
        """
        Add a panel to the network with stiffness k

        Parameters:
          k:      either a float stiffness or "rigid" or "disconnect"

        Returns:
          id of the new panel
        """
        name = self._panel_name(self.cpanel)
        self.cpanel += 1
        self.add_node(name, distance=1)
        self.add_edge("top", name, stiffness=k)

        return self.cpanel - 1

    def add_tube(self, panel, kconnect, k):
        """
        Add a tube to a panel

        Parameters:
          panel:      panel id
          kconnect:   either a float stiffness or "rigid" or "disconnect"
          k:          float stiffness
        """
        if panel >= self.cpanel:
            raise ValueError("Panel %i does not exist!" % panel)

        name = self._tube_name(panel, self._next_tube(panel))
        self.add_node(name + "_top", distance=2)
        self.add_node(name + "_bottom", distance=3)

        self.add_edge(self._panel_name(panel), name + "_top", stiffness=kconnect)
        self.add_edge(name + "_top", name + "_bottom", stiffness=k)

    def simplify_network(self):
        """
        Simplify the tube network and return a list of subproblems
        """
        # Remove all rigid nodes
        self._remove_rigid()

        # Split at all disconnects
        return self._split_disconnect()

    def solve_network(self, tforce):
        """
        Impose the standard BCs and solve for the displacements at each node

        Parameters:
          tforce:     thermal force on the bars
        """
        n = len(self)
        K = np.zeros((n, n))
        d = np.zeros((n,))
        f = np.zeros((n,))

        # Make the dof maps
        self.backmap = {}
        for i, name in enumerate(self.nodes()):
            self.backmap[name] = i

        # Fix all the ones with "_bottom" in the name to zero
        fixed = []
        for i, name in enumerate(self.nodes()):
            if name[:4] == "tube" and name.split("_")[1] == "bottom":
                fixed.append(i)
        free = sorted(list(set(list(range(n))) - set(fixed)))

        # Impose the thermal force and the stiffness matrix
        for a, b, data in self.edges(data=True):
            dofs = [self.backmap[a], self.backmap[b]]
            s = np.sign(self.nodes[b]["distance"] - self.nodes[a]["distance"])
            K[np.ix_(dofs, dofs)] += data["stiffness"] * np.array([[1, -1], [-1, 1]])

            if a[:4] == "tube" and b[:4] == "tube":
                f[dofs] += tforce * np.array([1, -1]) * s

        # Solve
        fprime = f[free] - np.dot(K[np.ix_(free, fixed)], d[fixed])
        Kprime = K[np.ix_(free, free)]
        dprime = la.solve(Kprime, fprime)
        d[free] = dprime

        # Backsub the displacements
        for i, (name, data) in enumerate(self.nodes(data=True)):
            data["displacement"] = d[i]

    def _remove_rigid(self):
        """
        Remove all rigid connections by deleting the edge and joining the
        nodes
        """
        while True:
            for a, b, key, data in self.edges(data=True, keys=True):
                if data["stiffness"] == "rigid":
                    self.remove_edge(a, b, key=key)
                    # Need to keep the one marked "tube"
                    if a[:4] == "tube":
                        ng = nx.contracted_nodes(self, a, b)
                    else:
                        ng = nx.contracted_nodes(self, b, a)
                    self.clear()
                    self.update(ng)
                    break
            else:
                break

    def _split_disconnect(self):
        """
        Split the graph at all disconnects
        """
        # First just split
        while True:
            for i, j, key, data in self.edges(keys=True, data=True):
                if data["stiffness"] == "disconnect":
                    self.remove_edge(i, j, key=key)
                    break
            else:
                break

        # Setup the new structures
        base = [
            Network(self.subgraph(c))
            for c in nx.connected_components(self)
            if len(c) > 1
        ]
        nstructs = []
        for b in base:
            ns = [n[:4] for n in b.nodes()]
            if "tube" in ns:
                nstructs.append(b)

        return nstructs

    def _panel_name(self, k):
        """
        Turn an integer into a panel id

        Parameters:
          k:      panel integer id
        """
        return "panel-%i" % k

    def _tube_name(self, panel, tube):
        """
        Turn integer panel and tube ids into a base tube string id
        """
        return "tube-%i-%i" % (panel, tube)

    def _next_tube(self, panel):
        """
        Figure out what the next tube id is for a panel
        """
        start = 0
        for n in self.neighbors(self._panel_name(panel)):
            if n != "top":
                start += 1

        return start


def gen_tube(k, tforce, tube_length, tube_ri, E, alpha, nr=5):
    emodel = elasticity.IsotropicLinearElasticModel(E, "youngs", 0.25, "poissons")
    mmodel = models.SmallStrainElasticity(emodel, alpha=alpha)

    solver = structural.PythonTubeSolver(atol=1.0e-4)

    ro = np.sqrt(tube_length * k / (np.pi * E) + tube_ri**2.0)
    A = np.pi * (ro**2.0 - tube_ri**2.0)

    tube = receiver.Tube(ro, ro - tube_ri, tube_length, nr, 1, 1)
    tube.set_times(np.array([0, 1.0]))
    tube.make_1D(tube_length / 2.0, 0)

    dT = tforce / (E * alpha * A)

    temp = np.zeros((2, nr))
    temp[1] = dT

    tube.add_results("temperature", temp)
    tube.set_pressure_bc(receiver.PressureBC(np.array([0, 1]), np.array([0, 0])))

    return spring.TubeSpring(tube, solver, mmodel)


def copy_network(base, tube_length, tube_ri, E, alpha, tforce):
    other = spring.SpringNetwork()

    for n in base.nodes():
        other.add_node(n)

    for a, b, data in base.edges(data=True):
        if a[:4] == "tube" and b[:4] == "tube":
            other.add_edge(
                a,
                b,
                object=gen_tube(
                    data["stiffness"], tforce, tube_length, tube_ri, E, alpha
                ),
            )
        else:
            if data["stiffness"] in ["rigid", "disconnect"]:
                other.add_edge(a, b, object=data["stiffness"])
            else:
                other.add_edge(a, b, object=spring.LinearSpring(data["stiffness"]))

    for n in other.nodes():
        if n[:4] == "tube" and n.split("_")[1] == "bottom":
            other.displacement_bc(n, lambda x: 0)

    other.set_times([0, 1])
    other.validate_setup()

    return other


class TestNetworksWithDisconnects(unittest.TestCase):
    def setUp(self):
        self.ntrials = 4
        self.tforce = 1.2

        self.kpc = 10.0
        self.ktc = 20.0
        self.kt = 5.0

        self.npanels = 5
        self.ntubes = 5

        self.tube_length = 1.0
        self.tube_ri = 0.9
        self.E = 120000.0
        self.alpha = 1.0e-5

    def test_random(self):
        ran = 0

        while ran < self.ntrials:
            ref, main = self.gen_network()

            subprobs1 = ref.simplify_network()
            subprobs1.sort(key=lambda n: len(n))
            subprobs2 = main.reduce_graph()
            subprobs2.sort(key=lambda n: len(n))

            sz1 = [len(n) for n in subprobs1]
            sz2 = [len(n) for n in subprobs2]

            b1 = [n for n in sz1 if n > 2]

            if len(b1) > 1:
                for i, j in zip(b1[1:], b1[:-1]):
                    if i == j:
                        break
                else:
                    continue

            for prob in subprobs1:
                prob.solve_network(self.tforce)

            for prob in subprobs2:
                prob.solve_all()

            for p1, p2 in zip(subprobs1, subprobs2):
                disp1 = np.sort(
                    np.array([data["displacement"] for n, data in p1.nodes(data=True)])
                )
                disp2 = np.sort(p2.displacements)

                self.assertEqual(len(disp1), len(disp2))
                print(disp1, disp2)
                self.assertTrue(np.allclose(disp1, disp2))

            ran += 1

    def gen_network(self):
        test = Network()

        for i in range(self.npanels):
            pid = test.add_panel(
                ra.choice(
                    [self.kpc, 0.8 * self.kpc, 1.2 * self.kpc] * 3
                    + ["rigid", "disconnect"]
                )
            )
            for j in range(self.ntubes):
                test.add_tube(
                    pid,
                    ra.choice(
                        [self.ktc, 0.8 * self.ktc, 1.2 * self.ktc] * 3
                        + ["rigid", "disconnect"]
                    ),
                    self.kt,
                )

        nn = copy_network(
            test, self.tube_length, self.tube_ri, self.E, self.alpha, self.tforce
        )
        old2new = {}
        new2old = {}
        for i, n in enumerate(nn.nodes()):
            old2new[n] = i
            new2old[i] = n
        nx.relabel_nodes(nn, old2new, copy=False)

        return test, nn


class TestNetworksNoDisconnects(unittest.TestCase):
    def setUp(self):
        self.ntrials = 4
        self.tforce = 1.2

        self.kpc = 10.0
        self.ktc = 20.0
        self.kt = 5.0

        self.npanels = 5
        self.ntubes = 5

        self.tube_length = 1.0
        self.tube_ri = 0.9
        self.E = 120000.0
        self.alpha = 1.0e-5

    def test_random(self):
        ran = 0

        while ran < self.ntrials:
            ref, main = self.gen_network()

            subprobs1 = ref.simplify_network()
            subprobs1.sort(key=lambda n: len(n))
            subprobs2 = main.reduce_graph()
            subprobs2.sort(key=lambda n: len(n))

            sz1 = [len(n) for n in subprobs1]
            sz2 = [len(n) for n in subprobs2]

            b1 = [n for n in sz1 if n > 2]

            if len(b1) > 1:
                for i, j in zip(b1[1:], b1[:-1]):
                    if i == j:
                        break
                else:
                    continue

            for prob in subprobs1:
                prob.solve_network(self.tforce)

            for prob in subprobs2:
                prob.solve_all()

            for p1, p2 in zip(subprobs1, subprobs2):
                disp1 = np.sort(
                    np.array([data["displacement"] for n, data in p1.nodes(data=True)])
                )
                disp2 = np.sort(p2.displacements)

                self.assertEqual(len(disp1), len(disp2))
                print(disp1, disp2)
                self.assertTrue(np.allclose(disp1, disp2, rtol=1e-4))

            ran += 1

    def gen_network(self):
        test = Network()

        for i in range(self.npanels):
            pid = test.add_panel(
                ra.choice([self.kpc, 0.8 * self.kpc, 1.2 * self.kpc] * 3 + ["rigid"])
            )
            for j in range(self.ntubes):
                test.add_tube(
                    pid,
                    ra.choice(
                        [self.ktc, 0.8 * self.ktc, 1.2 * self.ktc] * 3 + ["rigid"]
                    ),
                    ra.choice([self.kt, 0.8 * self.kt, 1.2 * self.kt]),
                )

        nn = copy_network(
            test, self.tube_length, self.tube_ri, self.E, self.alpha, self.tforce
        )
        old2new = {}
        new2old = {}
        for i, n in enumerate(nn.nodes()):
            old2new[n] = i
            new2old[i] = n
        nx.relabel_nodes(nn, old2new, copy=False)

        return test, nn



import numpy as np


def differentiate(fn, x0, eps=1.5e-6):
    if np.isscalar(x0):
        sx = (1,)
    else:
        sx = x0.shape

    f0 = fn(x0)

    if np.isscalar(f0):
        sf = (1,)
    else:
        sf = f0.shape

    tshape = sf + sx
    rshape = sx + sf

    D = np.zeros(rshape)

    for index in np.ndindex(sx):
        if np.isscalar(x0):
            diff = np.abs(x0 * eps)
        else:
            diff = np.abs(x0[index] * eps)
        if diff < eps:
            diff = eps

        if np.isscalar(x0):
            xp = diff
        else:
            xp = np.zeros(sx)
            xp[index] = diff

        fp = fn(xp + x0)

        D[index] = (fp - f0) / diff

    Df = np.zeros(tshape)
    # Reverse
    for ind1 in np.ndindex(sx):
        for ind2 in np.ndindex(sf):
            Df[ind2 + ind1] = D[ind1 + ind2]

    if np.isscalar(x0):
        return Df[0]
    else:
        return Df



"""
  This module defines 1D, 2D, and 3D thermal solvers.
"""

import multiprocess

import numpy as np
import numpy.linalg as la
import scipy.sparse as sp
import scipy.sparse.linalg as sla

from srlife import receiver, solverparams
from srlife.thermohydraulics import flowpath


class TemperatureResetter:
    """
    Reset the tube temperatures to a fixed value every so often.
    """

    def __init__(self, trigger, vals):
        """
        Setup the resetter

        Parameters:
          trigger     function which triggers the reset
          vals        values to reset to
        """
        self.trigger = trigger
        self.vals = vals

    def apply(self, time, step, temps):
        """
        Actually make the modification

        Parameters:
          time        current time
          step        current step
          temps       current temperatures
        """
        if self.trigger(time):
            temps[step] = self.vals


class ReceiverResetter:
    """
    Reset the tube temperatures to a fixed value every so often.
    This object works on the entire receiver at once
    """

    def __init__(self, trigger):
        """
        Setup the resetter

        Parameters:
          trigger     function which triggers the reset
        """
        self.trigger = trigger

    def apply(self, time, step, model):
        """
        Actually make the modification

        Parameters:
          time        current time
          step        current step
          temps       current temperatures
        """
        if self.trigger(time):
            for tube in model.tubes:
                tube.quadrature_results["ghost_temperature"][step] = tube.T0


class ThermohydraulicsThermalSolver:
    """
    Solve the heat transfer problem by iterating between a simple
    1D thermal hydraulics model and a FiniteDifferenceImplicitThermalSolver.
    """

    def __init__(self, pset=solverparams.ParameterSet()):
        self.rtol = pset.get_default("rtol", 1.0e-6)
        self.atol = pset.get_default("atol", 1.0e-3)
        self.miter = pset.get_default("miter", 1000)
        self.verbose = pset.get_default("verbose", False)
        self.eps = pset.get_default("epsilon", 1.0e-10)

        self.solid_params = pset.get_default("solid", solverparams.ParameterSet())
        self.thermo_params = pset.get_default("fluid", solverparams.ParameterSet())

    def solve_receiver(
        self,
        model,
        solid_material,
        fluid_material,
        decorator=None,
        nthreads=1,
        resetters=None,
    ):
        """Solve the entire receiver by splitting into independent flow paths

        Args:
            model (Receiver): fully-populated receiver object
            solid_material (ThermalMaterial): solid thermal properties
            fluid_material (ThermalFluidMaterial): fluid thermal properties
            decorator (Optional[decorator]): progress decorator
            nthreads(Optional[int]): number of allowable threads to use
            resetters(Optional[list]): resetting objects to apply
        """
        if resetters is None:
            resetters = []

        # Setup
        self.receiver = model
        self.solid_material = solid_material
        self.fluid_material = fluid_material
        self.decorator = decorator
        self.nthreads = nthreads

        # Stupid algorithm to decide on times
        solve_times = self._determine_solve_times()

        # Add the result fields we'll need, specifically nodal temperatures
        # and axial fluid temperatures
        for tube in self.receiver.tubes:
            tube.set_times(solve_times)
            tube.add_blank_axial_results("fluid_temperature")
            tube.add_blank_axial_results("fluid_velocity")
            tube.add_blank_quadrature_results(
                "ghost_temperature", (tube.ntime,) + tube_dim(tube)
            )

        # Check that we have enough information to solve the problem and
        # setup the initial conditions for metal and fluid temperature
        self._setup()

        # Loop over time and solve each time step
        def do_stuff(x):
            i = x[0] + 1
            time = x[1][0]
            dt = x[1][1]
            self.solve_step(i, time, dt)

            for resetter in resetters:
                resetter.apply(time, i, self.receiver)

        list(
            decorator(
                map(
                    do_stuff,
                    enumerate(zip(solve_times[1:], solve_times[1:] - solve_times[:-1])),
                ),
                len(solve_times) - 1,
            )
        )

        # Add the unghosted temperature field
        for tube in self.receiver.tubes:
            tube.add_results("temperature", self._unghost(tube))

    def _unghost(self, tube):
        """Remove the ghost nodes from tube "ghost_temperature" results"""
        T = tube.quadrature_results["ghost_temperature"]
        # Don't return ghost values
        if tube.abstraction == "3D":
            return T[:, 1:-1, 1:-1, 1:-1]
        elif tube.abstraction == "2D":
            return T[:, 1:-1, 1:-1]
        elif tube.abstraction == "1D":
            return T[:, 1:-1]
        else:
            raise ValueError("Unknown abstraction %s" % tube.abstraction)

    def solve_step(self, i, time, dt):
        """Solve timestep i at time time

        Args:
            i (int):        time step
            time (float):   actual time value
            dt (float):     time increment
        """
        # Setup initial guesses for the metal and tube temperatures (use last step!)
        for tube in self.receiver.tubes:
            tube.quadrature_results["ghost_temperature"][i] = tube.quadrature_results[
                "ghost_temperature"
            ][i - 1]
            tube.axial_results["fluid_temperature"][i] = tube.axial_results[
                "fluid_temperature"
            ][i - 1]
            tube.axial_results["fluid_velocity"][i] = tube.axial_results[
                "fluid_velocity"
            ][i - 1]

        # Iterate between solving the metal temperatures and solving the
        # fluid temperatures until both are fairly stationary
        if self.verbose:
            print("Solving timestep %i, discrete time %f" % (i, time))
        for j in range(self.miter):
            previous_temps = np.array(
                [
                    tube.quadrature_results["ghost_temperature"][i]
                    for panel in self.receiver.panels.values()
                    for tube in panel.tubes.values()
                ]
            )
            self.solve_metal(i, time, dt)
            next_temps = np.array(
                [
                    tube.quadrature_results["ghost_temperature"][i]
                    for panel in self.receiver.panels.values()
                    for tube in panel.tubes.values()
                ]
            )
            temp_diff = np.abs(next_temps - previous_temps)
            temp_max_diff = np.max(temp_diff)
            temp_max_rel_diff = np.max(temp_diff / (previous_temps + self.eps))
            if self.verbose:
                print(
                    "\tSolved metal temperatures for iteration %i, "
                    "absolute difference %e / relative difference %e"
                    % (j, temp_max_diff, temp_max_rel_diff)
                )

            previous_fluid_temps = np.array(
                [
                    tube.axial_results["fluid_temperature"][i]
                    for panel in self.receiver.panels.values()
                    for tube in panel.tubes.values()
                ]
            )
            self.solve_fluid(i, time, dt)
            next_fluid_temps = np.array(
                [
                    tube.axial_results["fluid_temperature"][i]
                    for panel in self.receiver.panels.values()
                    for tube in panel.tubes.values()
                ]
            )
            fluid_diff = np.abs(next_fluid_temps - previous_fluid_temps)
            fluid_max_diff = np.max(fluid_diff)
            fluid_max_rel_diff = np.max(fluid_diff / (previous_fluid_temps + self.eps))
            if self.verbose:
                print(
                    "\tSolved fluid temperatures for iteration %i, "
                    "absolute difference %e / relative difference %e"
                    % (j, fluid_max_diff, fluid_max_rel_diff)
                )

            if (fluid_max_diff < self.atol and temp_max_diff < self.atol) or (
                fluid_max_rel_diff < self.rtol and temp_max_rel_diff < self.rtol
            ):
                break
        else:
            raise RuntimeError(
                "Picard iteration in the thermohydraulics solver did not converge"
            )

    def solve_metal(self, i, time, dt):
        """Setup and solve for the metal temperatures in each tube"""
        # pylint: disable=no-member
        # Setup the appropriate convective BCs on the ID of each tube
        # This is the best use of threads as best I can tell
        for tube in self.receiver.tubes:
            tube.set_bc(
                receiver.FilmCoefficientConvectiveBC(
                    tube.r - tube.t,
                    tube.h,
                    tube.nz,
                    tube.axial_results["fluid_temperature"][i],
                    self.fluid_material.film_coefficient(
                        tube.axial_results["fluid_temperature"][i],
                        tube.axial_results["fluid_velocity"][i],
                        tube.r - tube.t,
                    ),
                ),
                "inner",
            )

        def work(tube):
            # Need to add params
            tube_solver = FiniteDifferenceImplicitThermalProblem(
                tube,
                self.solid_material,
                self.fluid_material,
                **deparametrize_finite_difference(self.solid_params)
            )
            # Solve...
            return tube_solver.solve_step_substep(
                tube.quadrature_results["ghost_temperature"][i - 1], time, dt
            )

        with multiprocess.Pool(self.nthreads) as p:
            data = list(p.map(work, self.receiver.tubes))

        for tube, res in zip(self.receiver.tubes, data):
            tube.quadrature_results["ghost_temperature"][i] = res

    def solve_fluid(self, i, time, dt):
        """Setup and solve for the fluid temperatures in each tube"""
        # Setup the thermohydraulic solver for each flow path
        for path in self.receiver.flowpaths.values():
            # Add parameters
            model = flowpath.FlowPath(
                path["times"],
                path["mass_flow"],
                path["inlet_temp"],
                **flowpath.deparameterize_flow_path(self.thermo_params)
            )
            for panel_name in path["panels"]:
                model.add_panel_from_object(
                    self.receiver.panels[panel_name], self.fluid_material
                )

            # Solve for the fluid temperatures
            nodal_temps = model.solve(time)

            # Update the stored fluid temperature and flow velocities
            flow_rates, tube_temperatures = model.recover_tube_results(
                nodal_temps, time
            )

            for panel_name, panel_rates, panel_temps in zip(
                path["panels"], flow_rates, tube_temperatures
            ):
                for k, tube in enumerate(
                    self.receiver.panels[panel_name].tubes.values()
                ):
                    tube.axial_results["fluid_temperature"][i] = panel_temps[k]
                    tube.axial_results["fluid_velocity"][i] = panel_rates[k]

    def _setup(self):
        """Setup for solve

        Sets initial conditions on the metal and fluid temperature and does some
        checking to make sure the problem is fully-defined
        """
        # Check that every panel is in exactly 1 flow path
        panels_in_path = []
        for path in self.receiver.flowpaths.values():
            panels_in_path.extend(path["panels"])
        pset = set(panels_in_path)
        if len(pset) != len(panels_in_path):
            raise ValueError("There are panels in more than one flow path!")
        if pset != set(self.receiver.panels.keys()):
            raise ValueError("At least one panel is not in a flow path!")

        # Setup each tube
        for path in self.receiver.flowpaths.values():
            for panel_name in path["panels"]:
                panel = self.receiver.panels[panel_name]
                for tube in panel.tubes.values():
                    tube.quadrature_results["ghost_temperature"][0] = tube.T0
                    tube.axial_results["fluid_temperature"][0] = path["inlet_temp"][0]
                    tube.axial_results["fluid_velocity"][
                        0
                    ] = 100000000.0  # Need to provide IC?

    def _determine_solve_times(self):
        """Determine which times to solve the temperatures for

        Dumb algorithm to determine times to solve at -- just look at a tube
        """
        return list(list(self.receiver.panels.values())[0].tubes.values())[0].times


def tube_dim(tube):
    """
    Figure out a tube's nodal field array size based on the discertization and
    abstraction
    """
    dim = (tube.nr + 2, tube.nt + 2, tube.nz + 2)
    if tube.abstraction == "1D":
        return dim[:1]
    elif tube.abstraction == "2D":
        return dim[:2]
    elif tube.abstraction == "3D":
        return dim
    else:
        raise ValueError("Unknown abstraction %s" % tube.abstraction)


def deparametrize_finite_difference(pset):
    """
    Convert a ParameterSet to kwargs with defaults
    """
    return {
        "rtol": pset.get_default("rtol", 1.0e-6),
        "atol": pset.get_default("atol", 1.0e-2),
        "miter": pset.get_default("miter", 100),
        "substep": pset.get_default("substep", 1),
        "verbose": pset.get_default("verbose", False),
        "steady": pset.get_default("steady", False),
    }


class FiniteDifferenceImplicitThermalSolver:
    """
    Solves the heat transfer problem using the finite difference method.

    Solver handles the *cylindrical* 1D, 2D, or 3D cases.
    """

    def __init__(
        self,
        pset=solverparams.ParameterSet(),
        rtol=1.0e-6,
        atol=1.0e-2,
        miter=100,
        substep=1,
        verbose=False,
        steady=False,
    ):
        """
        Setup the solver

        Additional parameters:
          pset        object with solver parameters
          rtol        iteration relative tolerance
          atol        iteration absolute tolerance
          miter       maximum iterations
          substep     divide user-provided time increments into smaller values
          verbose     print a lot of debug info
          steady      ignore thermal mass and use conduction only
        """
        self.rtol = pset.get_default("rtol", rtol)
        self.atol = pset.get_default("atol", atol)
        self.miter = pset.get_default("miter", miter)
        self.substep = pset.get_default("substep", substep)
        self.verbose = pset.get_default("verbose", verbose)
        self.steady = pset.get_default("steady", steady)

    def solve(
        self,
        tube,
        material,
        fluid,
        source=None,
        T0=None,
        fix_edge=None,
        rtol=1e-6,
        atol=1e-2,
        miter=100,
        substep=1,
        resetters=None,
    ):
        """
        Solve the thermal problem defined for a single tube

        Parameters:
          tube        the Tube object defining the geometry, loading,
                      and analysis abstraction
          material    the thermal material object describing the material
                      conductivity and diffusivity
          fluid       the fluid material object describing the convective
                      heat transfer coefficient

        Other Parameters:
          source      if present, the source term as a function of t and
                      then the coordinates
          T0          if present override the tube IC with a function of
                      the coordinates
          fix_edge    an exact solution to fix edge BCs for testing
          rtol        solver relative tolerance
          atol        solver absolute tolerance
          miter       maximum number of nonlinear iterations
          substep     subdivide thermal steps into smaller increments
          resetters   list of reset objects to apply
        """
        if resetters is None:
            resetters = []

        temperatures = FiniteDifferenceImplicitThermalProblem(
            tube,
            material,
            fluid,
            source,
            T0,
            fix_edge,
            self.rtol,
            self.atol,
            self.miter,
            self.substep,
            self.verbose,
            self.steady,
        ).solve(resetters)

        tube.add_results("temperature", temperatures)

        return temperatures


class FiniteDifferenceImplicitThermalProblem:
    """
    The actual finite difference solver created to solve a single
    tube problem
    """

    def __init__(
        self,
        tube,
        material,
        fluid,
        source=None,
        T0=None,
        fix_edge=None,
        rtol=1.0e-4,
        atol=1e-2,
        miter=50,
        substep=1,
        verbose=False,
        steady=False,
    ):
        """
        Parameters:
          tube        Tube object to solve
          material    ThermalMaterial object
          fluid       FluidMaterial object

        Other Parameters:
          source      source function (t, r, ...)
          T0          initial condition function
          fix_edge:   an exact solution to fix edge BCs for testing
          rtol        relative tolerance
          atol        absolute tolerance
          miter       maximum iterations
          substep     divide user provided time increments into smaller steps
          verbose     print a lot of debug info
          steady      use steady state (conduction) theory
        """
        self.tube = tube
        self.material = material
        self.fluid = fluid

        self.rtol = rtol
        self.atol = atol
        self.miter = miter

        self.substep = substep

        self.verbose = verbose

        self.steady = steady

        self.source_term = source
        self.T0 = T0
        self.fix_edge = fix_edge

        self.dr = self.tube.t / (self.tube.nr - 1)
        self.dt = 2.0 * np.pi / (self.tube.nt)
        self.dz = self.tube.h / (self.tube.nz - 1)

        self.dts = np.diff(self.tube.times)

        # Ghost
        self.dim = (self.tube.nr + 2, self.tube.nt + 2, self.tube.nz + 2)

        if self.tube.abstraction == "3D":
            self.dim = self.dim
            self.nr, self.nt, self.nz = self.dim
            self.ndim = 3
            self.fdim = self.dim
        elif self.tube.abstraction == "2D":
            self.dim = self.dim[:2]
            self.nr, self.nt = self.dim
            self.nz = 1
            self.ndim = 2
            self.fdim = (self.nr, self.nt, 1)
        elif self.tube.abstraction == "1D":
            self.dim = self.dim[:1]
            (self.nr,) = self.dim
            self.nt = 1
            self.nz = 1
            self.ndim = 1
            self.fdim = (self.nr, 1, 1)
        else:
            raise ValueError(
                "Thermal solver does not know how to handle"
                " abstraction %s" % self.tube.abstraction
            )

        # Useful for later
        self.mesh = self._generate_mesh()
        self.r = self.mesh[0].reshape(self.fdim)

        if self.ndim > 1:
            self.theta = self.mesh[1].reshape(self.fdim)
        else:
            self.theta = np.ones(self.r.shape) * self.tube.angle

        if self.ndim > 2:
            self.z = self.mesh[2].reshape(self.fdim)
        else:
            self.z = np.ones(self.r.shape) * self.tube.plane

    def _generate_mesh(self):
        """
        Produce the r, theta, z mesh
        """
        rs = np.linspace(
            self.tube.r - self.tube.t - self.dr, self.tube.r + self.dr, self.tube.nr + 2
        )
        ts = np.linspace(-self.dt, 2.0 * np.pi, self.tube.nt + 2)
        zs = np.linspace(0 - self.dz, self.tube.h + self.dz, self.tube.nz + 2)

        geom = [rs, ts, zs]

        return np.meshgrid(*geom[: self.ndim], indexing="ij", copy=True)

    def solve(self, resetters=None):
        """
        Actually solve the problem...
        """
        if resetters is None:
            resetters = []

        # Setup the initial time
        T = np.zeros((self.tube.ntime,) + self.dim)
        if self.T0 is not None:
            T[0] = self.T0(*self.mesh)
        else:
            T[0] = self.tube.T0

        # Iterate through steps
        for i, (time, dt) in enumerate(zip(self.tube.times[1:], self.dts)):
            T[i + 1] = self.solve_step_substep(T[i], time, dt)
            for r in resetters:
                r.apply(time, i + 1, T)

        return self._real_results(T)

    def _real_results(self, T):
        """Helper that only returns the actual results, without the ghosts"""
        # Don't return ghost values
        if self.ndim == 3:
            return T[..., 1:-1, 1:-1, 1:-1]
        elif self.ndim == 2:
            return T[..., 1:-1, 1:-1]
        else:
            return T[..., 1:-1]

    def solve_step_substep(self, T_n, time, dt):
        """
        Do substepping, if requested by the user

        Parameters:
          T_n         previous full step temperature
          time        target time
          dt          target dt
        """
        T = np.copy(T_n)
        t_n = time - dt
        dti = dt / self.substep

        for i in range(1, self.substep + 1):
            t = t_n + dti * i
            T = self.solve_step(T, t, dti)

        return T

    def setup_step(self, T):
        """
        Setup common reusable arrays

        Parameters:
          T           temperature of interest
          time        current time
          dt          current dt
        """
        self.k = self.material.conductivity(T).reshape(self.fdim)
        self.a = self.material.diffusivity(T).reshape(self.fdim)

        if self.steady:
            self.c = self.k
            self.qc = 1.0
        else:
            self.c = self.a
            self.qc = self.a / self.k

        # Useful matrix that gives you the actual dofs
        self.act = np.pad(
            np.ones(tuple(d - 2 for d in self.dim)),
            [(1, 1)] * self.ndim,
            constant_values=[(0, 0)] * self.ndim,
        )

        self.act = self.act.reshape(self.fdim)

        self.ndof = T.size

    def _generate_A(self):
        """
        Generate the base differential operator for the current step
        """
        A = self.radial()

        if self.ndim > 1:
            A += self.circumfrential()

        if self.ndim > 2:
            A += self.axial()

        return A

    def _generate_id(self):
        """
        Return the non-boundary affecting ID matrix
        """
        return sp.diags(
            [self.act.flatten()],
            offsets=(0,),
            shape=(self.ndof, self.ndof),
            format="coo",
        )

    def _generate_source(self, time):
        """
        Generate the source part of the RHS

        Parameters:
          time        current time
        """
        if self.source_term:
            return (
                self.act
                * self.qc
                * self.source_term(time, *[self.r, self.theta, self.z][: self.ndim])
            ).flatten()
        else:
            return np.zeros((self.ndof,))

    def _generate_prev_temp(self, T_n):
        """
        The RHS vector with the temperature contributions in the correct locations

        Parmaeters:
          T_n         previous temperatures
        """
        return (T_n * self.act).flatten()

    def _generate_bc_matrix(self, T_n, time):
        """
        Generate the BC matrix terms (fixed)
        """
        M = self._ID_BC() + self._OD_BC()

        if self.ndim > 1:
            M += self._left_BC() + self._right_BC()

        if self.ndim > 2:
            M += self._top_BC() + self._bot_BC()

        return M

    def _generate_fixed_bc_RHS(self, T_n, time):
        """
        Generate the constant (i.e. axial) contributions to the RHS

        Parameters:

        T_n       previous temperature
        time      current time
        dt        time increment
        """
        if self.ndim > 2:
            return self._top_BC_R(T_n, time, T_n) + self._bot_BC_R(T_n, time, T_n)
        else:
            return np.zeros((self.ndof,))

    # pylint: disable=too-many-locals
    def solve_step(self, T_n, time, dt):
        """
        Actually setup and solve for a single load step

        Parameters:
          T_n         previous temperatures
          time        current time
          dt          time increment
        """
        # Generic setup
        self.setup_step(T_n)

        # Add dimensions, if necessary
        T_n = T_n.reshape(self.fdim)

        # FD contributions
        A = self._generate_A()
        # Identity
        ID = self._generate_id()
        # Source term
        S = self._generate_source(time)
        # Previous temp
        Tn = self._generate_prev_temp(T_n)

        # System without BCs
        if self.steady:
            M = -A
            R = S
        else:
            M = ID - A * dt
            R = S * dt + Tn

        # Matrix and fixed RHS boundary contributions
        B = self._generate_bc_matrix(T_n, time)
        M += B
        BRF = self._generate_fixed_bc_RHS(T_n, time)
        R += BRF

        # Dummy dofs
        D = self._generate_dummy_dofs()
        M += D

        # Covert over our fixed matrix
        M = M.tocsr()

        # This would be the iteration step
        T = np.copy(T_n)

        for i in range(self.miter):
            Ri = R + self._ID_BC_R(T, time, T_n) + self._OD_BC_R(T, time, T_n)
            J = self._d_ID_BC_R(T, time, T_n) + self._d_OD_BC_R(T, time, T_n)

            res = M.dot(T.flatten()) - Ri
            nr = la.norm(res)

            if i == 0:
                nr0 = nr

            if (nr < self.atol or nr / nr0 < self.rtol) and i > 0:
                break

            T -= sla.spsolve(M - J, res).reshape(self.fdim)
        else:
            raise RuntimeError("Too many iterations in newton solver!")

        return T.reshape(self.dim)

    # pylint: disable=too-many-branches
    def _generate_dummy_dofs(self):
        """
        Provide on-diagonal 1.0s for the dummy dofs
        """
        I = []
        J = []

        if self.ndim == 2:
            for i in self.dummy_loop_r():
                for j in self.dummy_loop_t():
                    for k in self.dummy_loop_z():
                        I.append(self.dof(i, j, k))
                        J.append(self.dof(i, j, k))
        elif self.ndim == 3:
            for i in self.dummy_loop_r():
                for j in self.dummy_loop_t():
                    for k in self.full_loop_z():
                        I.append(self.dof(i, j, k))
                        J.append(self.dof(i, j, k))
            for i in self.full_loop_r():
                for j in self.dummy_loop_t():
                    for k in self.dummy_loop_z():
                        I.append(self.dof(i, j, k))
                        J.append(self.dof(i, j, k))
            for i in self.dummy_loop_r():
                for j in self.full_loop_t():
                    for k in self.dummy_loop_z():
                        I.append(self.dof(i, j, k))
                        J.append(self.dof(i, j, k))

        return sp.coo_matrix((np.ones((len(I),)), (I, J)), shape=(self.ndof, self.ndof))

    def _ID_BC_R(self, T, time, T_n):
        """
        The inner diameter BC RHS contribution

        Parameters:
          T       current temperatures
          time    current time
          T_n     previous temperatures
        """
        R = np.zeros((self.ndof,))
        i = 0
        for j in self.loop_t():
            for k in self.loop_z():
                if self.fix_edge:
                    R[self.dof(i, j, k)] = self.fix_edge(
                        time,
                        *[self.r[i, j, k], self.theta[i, j, k], self.z[i, j, k]][
                            : self.ndim
                        ]
                    )
                # Zero flux
                elif self.tube.inner_bc is None:
                    R[self.dof(i, j, k)] = 0.0
                # Fixed temperature
                elif isinstance(self.tube.inner_bc, receiver.FixedTempBC):
                    R[self.dof(i, j, k)] = self.tube.inner_bc.temperature(
                        time, self.theta[1, j, k], self.z[1, j, k]
                    )
                # Fixed flux
                elif isinstance(self.tube.inner_bc, receiver.HeatFluxBC):
                    R[self.dof(i, j, k)] = (
                        -self.dr
                        * self.tube.inner_bc.flux(
                            time, self.theta[1, j, k], self.z[1, j, k]
                        )
                        / self.k[1, j, k]
                    )
                # Convection
                elif isinstance(self.tube.inner_bc, receiver.ConvectiveBC):
                    fluid_T = self.tube.inner_bc.fluid_temperature(
                        time, self.z[1, j, k]
                    )[0]
                    R[self.dof(i, j, k)] = (
                        self.dr
                        * self.fluid.coefficient(self.material.name, fluid_T)
                        * (T[1, j, k] - fluid_T)
                        / self.k[1, j, k]
                    )
                # Convection giving the film coefficient and fluid temperature
                elif isinstance(
                    self.tube.inner_bc, receiver.FilmCoefficientConvectiveBC
                ):
                    fluid_T = self.tube.inner_bc.fluid_temperature(
                        time, self.z[1, j, k]
                    )
                    h = self.tube.inner_bc.film_coefficient(time, self.z[1, j, k])
                    R[self.dof(i, j, k)] = (
                        self.dr * h * (T[1, j, k] - fluid_T) / self.k[1, j, k]
                    )
                else:
                    raise ValueError("Unknown boundary condition!")
        return R

    def _d_ID_BC_R(self, T, time, T_n):
        """
        Derivative of the inner diameter BC RHS contribution

        Parameters:
          T       current temperatures
          time    current time
          T_n     previous temperatures
        """
        I = []
        J = []
        D = []
        if isinstance(self.tube.inner_bc, receiver.ConvectiveBC):
            i = 0
            for j in self.loop_t():
                for k in self.loop_z():
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(1, j, k))
                    fluid_T = self.tube.inner_bc.fluid_temperature(
                        time, self.z[1, j, k]
                    )[0]
                    D.append(
                        self.dr
                        * self.fluid.coefficient(self.material.name, fluid_T)
                        / self.k[1, j, k]
                    )
        elif isinstance(self.tube.inner_bc, receiver.FilmCoefficientConvectiveBC):
            i = 0
            for j in self.loop_t():
                for k in self.loop_z():
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(1, j, k))
                    fluid_T = self.tube.inner_bc.fluid_temperature(
                        time, self.z[1, j, k]
                    )
                    h = self.tube.inner_bc.film_coefficient(time, self.z[1, j, k])
                    D.append(self.dr * h / self.k[1, j, k])

        return sp.coo_matrix((D, (I, J)), shape=(self.ndof, self.ndof))

    def _OD_BC_R(self, T, time, T_n):
        """
        The outer diameter BC RHS contribution

        Parameters:
          T       current temperatures
          time    current time
          T_n     previous temperatures
        """
        R = np.zeros((self.ndof,))
        i = self.nr - 1
        for j in self.loop_t():
            for k in self.loop_z():
                if self.fix_edge:
                    R[self.dof(i, j, k)] = self.fix_edge(
                        time,
                        *[self.r[i, j, k], self.theta[i, j, k], self.z[i, j, k]][
                            : self.ndim
                        ]
                    )
                # Zero flux
                elif self.tube.outer_bc is None:
                    R[self.dof(i, j, k)] = 0.0
                # Fixed temperature
                elif isinstance(self.tube.outer_bc, receiver.FixedTempBC):
                    R[self.dof(i, j, k)] = self.tube.outer_bc.temperature(
                        time, self.theta[self.nr - 2, j, k], self.z[self.nr - 2, j, k]
                    )
                # Fixed flux
                elif isinstance(self.tube.outer_bc, receiver.HeatFluxBC):
                    R[self.dof(i, j, k)] = (
                        -self.dr
                        * self.tube.outer_bc.flux(
                            time,
                            self.theta[self.nr - 2, j, k],
                            self.z[self.nr - 2, j, k],
                        )
                        / self.k[self.nr - 2, j, k]
                    )
                # Convection
                elif isinstance(self.tube.outer_bc, receiver.ConvectiveBC):
                    fluid_T = self.tube.outer_bc.fluid_temperature(
                        time, self.z[self.nr - 2, j, k]
                    )[0]
                    R[self.dof(i, j, k)] = (
                        self.dr
                        * self.fluid.coefficient(self.material.name, fluid_T)
                        * (T[self.nr - 2, j, k] - fluid_T)
                        / self.k[self.nr - 2, j, k]
                    )
                else:
                    raise ValueError("Unknown boundary condition!")
        return R

    def _d_OD_BC_R(self, T, time, T_n):
        """
        Derivative of the outer diameter BC RHS contribution

        Parameters:
          T       current temperatures
          time    current time
          T_n     previous temperatures
        """
        I = []
        J = []
        D = []
        if isinstance(self.tube.outer_bc, receiver.ConvectiveBC):
            i = self.nr - 1
            for j in self.loop_t():
                for k in self.loop_z():
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(self.nr - 2, j, k))
                    fluid_T = self.tube.outer_bc.fluid_temperature(
                        time, self.z[self.nr - 2, j, k]
                    )[0]
                    D.append(
                        self.dr
                        * self.fluid.coefficient(self.material.name, fluid_T)
                        / self.k[self.nr - 2, j, k]
                    )

        return sp.coo_matrix((D, (I, J)), shape=(self.ndof, self.ndof))

    def _ID_BC(self):
        """
        Inner diameter boundary condition contribution matrix
        """
        I = []
        J = []
        D = []
        i = 0
        for j in self.loop_t():
            for k in self.loop_z():
                if self.fix_edge:
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(i, j, k))
                    D.append(1.0)
                # Zero flux
                elif self.tube.inner_bc is None:
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(1, j, k))
                    D.append(1.0)
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(0, j, k))
                    D.append(-1.0)
                # Fixed temperature
                elif isinstance(self.tube.inner_bc, receiver.FixedTempBC):
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(1, j, k))
                    D.append(1.0)
                # Fixed flux
                elif isinstance(self.tube.inner_bc, receiver.HeatFluxBC):
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(1, j, k))
                    D.append(1.0)
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(0, j, k))
                    D.append(-1.0)
                # Convection
                elif isinstance(
                    self.tube.inner_bc,
                    (receiver.ConvectiveBC, receiver.FilmCoefficientConvectiveBC),
                ):
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(1, j, k))
                    D.append(1.0)
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(0, j, k))
                    D.append(-1.0)
                else:
                    raise ValueError("Unknown boundary condition!")

        return sp.coo_matrix((D, (I, J)), shape=(self.ndof, self.ndof))

    def _OD_BC(self):
        """
        Outer diameter contribution to the BC matrix
        """
        I = []
        J = []
        D = []
        i = self.nr - 1
        for j in self.loop_t():
            for k in self.loop_z():
                if self.fix_edge:
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(i, j, k))
                    D.append(1.0)
                # Zero flux
                elif self.tube.outer_bc is None:
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(self.nr - 2, j, k))
                    D.append(1.0)

                    I.append(self.dof(i, j, k))
                    J.append(self.dof(self.nr - 1, j, k))
                    D.append(-1.0)
                # Fixed temperature
                elif isinstance(self.tube.outer_bc, receiver.FixedTempBC):
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(self.nr - 2, j, k))
                    D.append(1.0)
                # Fixed flux
                elif isinstance(self.tube.outer_bc, receiver.HeatFluxBC):
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(self.nr - 2, j, k))
                    D.append(1.0)
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(self.nr - 1, j, k))
                    D.append(-1.0)
                # Convection
                elif isinstance(self.tube.outer_bc, receiver.ConvectiveBC):
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(self.nr - 2, j, k))
                    D.append(1.0)
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(self.nr - 1, j, k))
                    D.append(-1.0)
                else:
                    raise ValueError("Unknown boundary condition!")

        return sp.coo_matrix((D, (I, J)), shape=(self.ndof, self.ndof))

    def _left_BC(self):
        """
        Periodic contribution to the BC matrix
        """
        I = []
        J = []
        D = []
        j = 0
        for i in self.loop_r():
            for k in self.loop_z():
                I.append(self.dof(i, j, k))
                J.append(self.dof(i, j, k))
                D.append(1.0)

                I.append(self.dof(i, j, k))
                J.append(self.dof(i, self.nt - 2, k))
                D.append(-1.0)

        return sp.coo_matrix((D, (I, J)), shape=(self.ndof, self.ndof))

    def _right_BC(self):
        """
        Periodic contribution to the BC matrix
        """
        I = []
        J = []
        D = []
        j = self.nt - 1
        for i in self.loop_r():
            for k in self.loop_z():
                I.append(self.dof(i, j, k))
                J.append(self.dof(i, j, k))
                D.append(1.0)

                I.append(self.dof(i, j, k))
                J.append(self.dof(i, 1, k))
                D.append(-1.0)

        return sp.coo_matrix((D, (I, J)), shape=(self.ndof, self.ndof))

    def _top_BC(self):
        """
        Axial top contribution to the BC matrix
        """
        I = []
        J = []
        D = []
        k = 0
        for i in self.loop_r():
            for j in self.loop_t():
                if self.fix_edge:
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(i, j, k))
                    D.append(1.0)
                else:
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(i, j, 1))
                    D.append(1.0)
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(i, j, 0))
                    D.append(-1.0)

        return sp.coo_matrix((D, (I, J)), shape=(self.ndof, self.ndof))

    def _bot_BC(self):
        """
        Axial bottom contribution to the BC matrix
        """
        I = []
        J = []
        D = []
        k = self.nz - 1
        for i in self.loop_r():
            for j in self.loop_t():
                if self.fix_edge:
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(i, j, k))
                    D.append(1.0)
                else:
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(i, j, self.nz - 2))
                    D.append(1.0)
                    I.append(self.dof(i, j, k))
                    J.append(self.dof(i, j, self.nz - 1))
                    D.append(-1.0)

        return sp.coo_matrix((D, (I, J)), shape=(self.ndof, self.ndof))

    def _top_BC_R(self, T, time, T_n):
        """
        RHS contribution of the top axial BC

        Parameters:
          T           current temperatures
          time        current time
          T_n         previous temperatures
        """
        R = np.zeros((self.ndof,))
        k = 0
        for i in self.loop_r():
            for j in self.loop_t():
                if self.fix_edge:
                    R[self.dof(i, j, k)] = self.fix_edge(
                        time,
                        *[self.r[i, j, k], self.theta[i, j, k], self.z[i, j, k]][
                            : self.ndim
                        ]
                    )

        return R

    def _bot_BC_R(self, T, time, T_n):
        """
        RHS contribution of the bottom axial BC

        Parameters:
          T           current temperatures
          time        current time
          T_n         previous temperatures
        """
        R = np.zeros((self.ndof,))
        k = self.nz - 1
        for i in self.loop_r():
            for j in self.loop_t():
                if self.fix_edge:
                    R[self.dof(i, j, k)] = self.fix_edge(
                        time,
                        *[self.r[i, j, k], self.theta[i, j, k], self.z[i, j, k]][
                            : self.ndim
                        ]
                    )

        return R

    def dof(self, i, j, k):
        """
        Return the DOF corresponding to the given grid position

        Parameters:
          i       r index
          j       theta index
          k       z index
        """
        return i * self.nt * self.nz + j * self.nz + k

    def loop_r(self):
        """
        Loop over non-ghost dofs
        """
        return range(1, self.nr - 1)

    def loop_t(self):
        """
        Loop over non-ghost dofs
        """
        if self.ndim > 1:
            return range(1, self.nt - 1)
        else:
            return [0]

    def loop_z(self):
        """
        Loop over non-ghost dofs
        """
        if self.ndim > 2:
            return range(1, self.nz - 1)
        else:
            return [0]

    def full_loop_r(self):
        """
        Loop over all dofs
        """
        return range(0, self.nr)

    def full_loop_t(self):
        """
        Loop over all dofs
        """
        if self.ndim > 1:
            return range(0, self.nt)
        else:
            return [0]

    def full_loop_z(self):
        """
        Loop over all dofs
        """
        if self.ndim > 2:
            return range(0, self.nz)
        else:
            return [0]

    def dummy_loop_r(self):
        """
        Loop over ghost dofs
        """
        return [0, self.nr - 1]

    def dummy_loop_t(self):
        """
        Loop over ghost dofs
        """
        return [0, self.nt - 1]

    def dummy_loop_z(self):
        """
        Loop over ghost dofs
        """
        if self.ndim > 2:
            return [0, self.nz - 1]
        else:
            return [0]

    def radial(self):
        """
        Insert the radial FD contribution into a sparse matrix
        """
        rh = (self.r[:-1] + self.r[1:]) / 2.0
        ah = (self.c[:-1] + self.c[1:]) / 2.0
        rhah = np.pad(rh * ah, ((1, 1), (0, 0), (0, 0)), mode="edge")

        D1 = (self.act * rhah[:-1] / (self.r * self.dr**2.0)).flatten()[
            self.nt * self.nz :
        ]
        D2 = -(self.act * (rhah[:-1] + rhah[1:]) / (self.r * self.dr**2.0)).flatten()
        D3 = (self.act * rhah[1:] / (self.r * self.dr**2.0)).flatten()[
            : -self.nt * self.nz
        ]

        return sp.diags(
            (D1, D2, D3),
            offsets=(-self.nt * self.nz, 0, self.nt * self.nz),
            shape=(self.ndof, self.ndof),
            format="coo",
        )

    def circumfrential(self):
        """
        Insert the circumferential FD contribution into a coo matrix
        """
        ah = np.pad(
            (self.c[:, :-1] + self.c[:, 1:]) / 2.0,
            ((0, 0), (1, 1), (0, 0)),
            mode="edge",
        )

        D1 = (self.act * ah[:, :-1] / (self.r**2.0 * self.dt**2.0)).flatten()[
            self.nz :
        ]
        D2 = -(
            self.act * (ah[:, :-1] + ah[:, 1:]) / (self.r**2.0 * self.dt**2.0)
        ).flatten()
        D3 = (self.act * ah[:, 1:] / (self.r**2.0 * self.dt**2.0)).flatten()[
            : -self.nz
        ]

        return sp.diags(
            (D1, D2, D3),
            offsets=(-self.nz, 0, self.nz),
            shape=(self.ndof, self.ndof),
            format="coo",
        )

    def axial(self):
        """
        Insert the axial FD contribution into a coo matrix
        """
        ah = np.pad(
            (self.c[:, :, :-1] + self.c[:, :, 1:]) / 2.0,
            ((0, 0), (0, 0), (1, 1)),
            mode="edge",
        )

        D1 = (self.act * ah[:, :, :-1] / self.dz**2.0).flatten()[1:]
        D2 = (-self.act * (ah[:, :, :-1] + ah[:, :, 1:]) / self.dz**2.0).flatten()
        D3 = (self.act * ah[:, :, 1:] / self.dz**2.0).flatten()[:-1]

        return sp.diags(
            [D1, D2, D3], offsets=(-1, 0, 1), shape=(self.ndof, self.ndof), format="coo"
        )



# pylint: disable=unnecessary-lambda

"""
  A very small module defining a class to help organize solver parameters
"""

from collections import defaultdict


class ParameterSet(defaultdict):
    """Recursive default dictionary that can be pickled

    Args:
      **kwargs: Arbitrary keyword arguments to be added to dictionary
    """

    def __init__(self, **kwargs):
        super().__init__(lambda: self.__class__(**kwargs))
        self.__dict__.update(kwargs)

    def __reduce__(self):
        """Reduce to something pickle-able"""
        return (type(self), (), None, None, iter(self.items()))

    def get_default(self, key, value):
        """Get a key, if not present return the default value"""
        if key in self.keys():
            return self[key]
        else:
            return value



#!/usr/bin/env python3

import os.path
import time

import numpy as np
import numpy.linalg as la
import scipy.io as sio

import matplotlib.pyplot as plt
from neml import models, elasticity, parse

import sys

sys.path.append("../../..")
from srlife import receiver, structural

# Moose comparison stuff
# MOOSE test stuff
ramp = lambda x: np.piecewise(
    x, [x < 1, x >= 1], [lambda xx: xx, lambda xx: 1.0 + xx * 0.0]
)
unit_temperature = (
    lambda x, y, z: (np.sqrt(x**2.0 + y**2.0) - 9)
    / (10 - 9)
    * np.cos(np.arctan2(y, x))
    * (z / 10.0 + 1)
)
temperature = lambda t, x, y, z: np.array(
    [100 * ramp(ti) * unit_temperature(x, y, z) for ti in t]
)
pressure = lambda t: 1.0 * ramp(t)

times = np.array([0, 0.5, 1, 101, 201, 301, 401, 501, 601, 701, 801, 901, 1001])
mat = parse.parse_xml(
    os.path.join(os.path.dirname(__file__), "moose-verification", "model.xml"),
    "creeping",
)

ri = 9.0
ro = 10.0
h = 10.0
nr = 11
nt = 20
nz = 6

moose_base = os.path.join(os.path.dirname(__file__), "moose-verification")
moose_ver = [os.path.join(moose_base, f) for f in ["1d_out.e", "2d_out.e", "3d_out.e"]]


def run_reference_simulation(d, solver):
    tube = receiver.Tube(ro, ro - ri, h, nr, nt, nz)
    tube.set_times(times)

    if d == 1:
        tube.make_1D(0, 0)
    elif d == 2:
        tube.make_2D(0)

    R, T, Z = tube.mesh
    X = R * np.cos(T)
    Y = R * np.sin(T)

    Ts = temperature(times, X, Y, Z).reshape((len(times),) + tube.dim[: tube.ndim])

    tube.add_results("temperature", Ts)
    tube.set_pressure_bc(receiver.PressureBC(times, pressure(times)))

    solver.setup_tube(tube)

    state_n = solver.init_state(tube, mat)

    for i in range(1, len(tube.times)):
        state_np1 = solver.solve(tube, i, state_n, 0.0)
        solver.dump_state(tube, i, state_np1)
        state_n = state_np1

    return tube


if __name__ == "__main__":
    if len(sys.argv) != 2:
        raise ValueError("Need to supply one argument, problem dimension!")

    d = int(sys.argv[1])

    if d not in [1, 2, 3]:
        raise ValueError("Invalid problem dimension %i!" % d)

    solver = structural.PythonTubeSolver()

    print("Running %iD problem..." % d)

    t1 = time.time()
    tube = run_reference_simulation(d, solver)
    t2 = time.time()

    print("Walltime: %f s" % (t2 - t1))



import unittest
import tempfile

from absl import logging

logging.set_verbosity(logging.FATAL)

import numpy as np
import numpy.random as ra
from srlife.thermohydraulics import thermalfluid, flowpath


class TestBasicSetup(unittest.TestCase):
    def setUp(self):
        self.cp = np.array([0.1046, 944.622])
        self.rho = np.array([-0.522, 1903.7])
        self.mu = np.array([1.784e-8, -2.91e-5, 1.4965e-2])
        self.k = np.array([-0.0001, 0.5047])

        self.material = thermalfluid.PolynomialThermalFluidMaterial(
            self.cp, self.rho, self.mu, self.k
        )

        self.times = np.array([0.0, 1.0, 2.0])
        self.mass_flow = np.array([10.0, 100.0, 10.0])
        self.weights = np.array([2.0, 4.0, 2.0])
        self.ri = 25.4
        self.h = 1000.0
        self.ntime = len(self.times)
        self.ntube = len(self.weights)
        self.nt = 12
        self.nz = 14
        self.metal_temp = ra.random((self.ntime, self.ntube, self.nt, self.nz))

        self.panel1 = flowpath.SimplePanelLink(
            self.times,
            self.mass_flow,
            self.weights,
            self.ri,
            self.h,
            self.metal_temp,
            self.material,
        )

        self.t = 1.2

    def test_basic_balance(self):
        T_start = np.array([100.0])
        T_end = np.array([120.0, 140.0, 110.0])

        Q_mass = self.panel1.Q_mass(T_start, T_end, self.t)
        self.assertEqual(Q_mass.shape, (self.ntube,))

        Q_conv = self.panel1.Q_conv(T_start, T_end, self.t)
        self.assertEqual(Q_conv.shape, (self.ntube,))

        R = self.panel1.residual(T_start, T_end, self.t)
        self.assertTrue(np.allclose(R, Q_mass - Q_conv))


class TestIsothermal(unittest.TestCase):
    def setUp(self):
        self.cp = np.array([994.6])
        self.rho = np.array([1903.7])
        self.mu = np.array([1.5e-2])
        self.k = np.array([0.5])

        self.material = thermalfluid.PolynomialThermalFluidMaterial(
            self.cp, self.rho, self.mu, self.k
        )

        self.times = np.array([0.0, 1.0])
        self.mass_flow = np.array([0.0, 100.0])
        self.inlet_temperature = np.array([0.0, 50.0])

        self.ntime = len(self.times)
        self.nt = 12
        self.nz = 10

        self.weights_1 = np.array([1.0, 1.0, 1.0])
        self.ri_1 = 25.4
        self.h_1 = 1000.0
        self.metal_1 = np.zeros((self.ntime, len(self.weights_1), self.nt, self.nz))

        self.metal_1[1, 0] = 50.0
        self.metal_1[1, 1] = 50.0
        self.metal_1[1, 2] = 50.0

        self.model = flowpath.FlowPath(
            self.times, self.mass_flow, self.inlet_temperature
        )
        self.model.add_panel(
            self.weights_1, self.ri_1, self.h_1, self.metal_1, self.material
        )

    def test_no_heat_in(self):
        T1 = self.model.solve(1.0)
        self.assertTrue(np.isclose(T1[-1], self.inlet_temperature[-1]))


class TestAnalytic(unittest.TestCase):
    def setUp(self):
        self.cp = np.array([994.6])
        self.rho = np.array([1.9])
        self.mu = np.array([1.5e-6])
        self.k = np.array([0.5])

        self.material = thermalfluid.PolynomialThermalFluidMaterial(
            self.cp, self.rho, self.mu, self.k
        )

        self.times = np.array([0.0, 1.0])
        self.mass_flow = np.array([0.0, 10.0])
        self.inlet_temperature = np.array([0.0, 50.0])

        self.ntime = len(self.times)
        self.nt = 12
        self.nz = 10

        self.weights_1 = np.array([2.0, 10.0, 2.0])
        self.ri_1 = 100.0
        self.h_1 = 1000.0
        self.metal_1 = np.zeros((self.ntime, len(self.weights_1), self.nt, self.nz))

        self.metal_T = 100.0

        self.metal_1[1, 0] = self.metal_T
        self.metal_1[1, 1] = self.metal_T
        self.metal_1[1, 2] = self.metal_T

        self.model = flowpath.FlowPath(
            self.times, self.mass_flow, self.inlet_temperature
        )
        self.model.add_panel(
            self.weights_1, self.ri_1, self.h_1, self.metal_1, self.material
        )

    def test_no_heat_in(self):
        T = self.model.solve(1.0)
        Tnumer = T[-1]
        u = self.mass_flow[1] / (
            self.rho * np.sum(self.weights_1) * np.pi * self.ri_1**2.0
        )
        h_film = self.material.film_coefficient(
            np.array([100.0]), u, np.array([self.ri_1])
        )[0]

        mdot = self.mass_flow[-1]
        cp = self.cp[0]
        Ts = self.inlet_temperature[-1]
        r = self.ri_1
        h = self.h_1
        Tm = self.metal_T
        n = np.sum(self.weights_1)

        Tanal = (
            mdot * cp * Ts + n * 2.0 * np.pi * r * h * h_film * (Tm - 0.5 * Ts)
        ) / (mdot * cp + n * 2.0 * np.pi * r * h * h_film * 0.5)

        self.assertAlmostEqual(Tanal, Tnumer)



# pylint: disable=wrong-import-position
"""
    Simple 1D heat balance thermohydraulics solver
"""

import numpy as np
import numpy.linalg as la
import scipy.interpolate as inter
import scipy.sparse as sp
import scipy.sparse.linalg as spla

from jax.config import config

config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax import jacfwd


class StartLink:
    """Starting link in the chain, gives the first temperature"""

    def __init__(self, times, inlet_temperature, **kwargs):
        """
        Parameters:
            times (ntime,):                 discrete times
            inlet_temperature (ntime,):     inlet temperatures
        """
        self.times = times
        self.inlet_temperature = inlet_temperature

        self.inlet_ifn = inter.interp1d(self.times, self.inlet_temperature)

        # Fixed
        self.size = 1

    def T_inlet(self, t):
        """Calculate the inlet temperature at a given time

        Parameters:
            t:      time
        """
        return jnp.asarray(self.inlet_ifn(t))

    def residual(self, T_start, T_end, t):
        """Calculate the residual contribution

        Parameters:
            T_start:    inlet temperatures (null here)
            T_end:      outlet temperatures
            t:          time
        """
        return T_end - self.T_inlet(t)


class PanelLink:
    """Connection representing part of a panel"""

    def __init__(self, times, mass_flow, weights, **kwargs):
        """
        Parameters:
            times (ntime,):     discrete times
            mass_flow (ntime,): discrete flow rates
            weights (ntube,):   number of actual tube represented by each panel
        """
        self.times = times
        self.mass_flow = mass_flow
        self.weights = jnp.asarray(weights)

        self.mass_ifn = inter.interp1d(self.times, self.mass_flow)

    def mass_flow_rate(self, t):
        """
        Calculate the mass flow at a given time
        """
        return jnp.asarray(self.mass_ifn(t))

    @property
    def ntube(self):
        """Number of actual tubes in the panel"""
        return jnp.sum(self.weights)


class ManifoldLink(PanelLink):
    """Averages temperatures together at this node in the flow path"""

    def __init__(self, times, mass_flow, weights, **kwargs):
        """
        Args:
            times (ntime,):     times for input
            mass_flow (ntime,): mass flow rates
            weights (ntube,):   tube weights
        """
        super().__init__(times, mass_flow, weights, **kwargs)

        # Fixed again...
        self.size = 1

    def residual(self, T_in, T_out, t):
        """Residual equations

        Parameters:
            T_in:       inlet temperatures
            T_out:      outlet temperatures
            t:          time
        """
        return jnp.sum(self.weights * T_in) / self.ntube - T_out


class SimplePanelLink(PanelLink):
    """Very simple panel model for debugging"""

    def __init__(
        self, times, mass_flow, weights, ri, h, metal_temp, material, **kwargs
    ):
        """
        Parameters:
            times (ntime,):                 times for input
            mass_flow (ntime,):             panel mass flow rate
            weights (ntube,):               tube weights
            ri:                             tube inner radius
            h:                              tube height
            metal_temp (ntime,ntube,nt,nz): tube metal temperatures, fixed grid
            material:                       thermofluid model
        """
        super().__init__(times, mass_flow, weights, **kwargs)
        self.ri = ri
        self.h = h
        self.metal_temp = metal_temp
        self.material = material
        self.zs = np.linspace(0, self.h, self.metal_temp.shape[3])
        self.dtheta = 2.0 * np.pi / self.metal_temp.shape[2]
        self.dz = self.h / self.metal_temp.shape[3]

        self.metal_ifn = inter.interp1d(self.times, self.metal_temp, axis=0)

    @property
    def size(self):
        """Number of explicitly-represented tubes"""
        return len(self.weights)

    def residual(self, T_in, T_out, t):
        """Residual equations

        Args:
            T_in:       inlet temperatures
            T_out:      outlet temperatures
            t:          time
        """
        return self.Q_mass(T_in, T_out, t) - self.Q_conv(T_in, T_out, t)

    def mean_temperature(self, T_start, T_tube):
        """
        Returns the mean temperature of each tube
        """
        return (T_tube + T_start) / 2.0

    def metal_temperature(self, t):
        """
        Metal temperature field at a given time
        """
        return jnp.asarray(self.metal_ifn(t))

    def Q_mass(self, T_start, T_tube, t):
        """
        Heat transfered out of the panel due to mass flow

        Parameters:
            T_start (scalar):   temperature at start of panel
            T_tube (ntube,):    temperature at end of tubes
            t:                  time
        """
        mdot = self.mass_flow_rate(t)
        T_mean = self.mean_temperature(T_start, T_tube)
        cp = self.material.cp(T_mean)

        return self.weights * mdot / self.ntube * cp * (T_tube - T_start)

    def Q_conv(self, T_start, T_tube, t):
        """
        Heat transfered into the panel due to convection

        Parameters:
            T_start (scalar):   temperature at start of panel
            T_tube (ntube,):    temperature at end of tubes
            t:                  time
        """
        # Flow velocity
        u = self.flow_rates(T_start, T_tube, t)

        # Convective heat transfer coefficient
        # We could make this height dependent if we're using AD...
        T_mean = self.mean_temperature(T_start, T_tube)
        h = self.material.film_coefficient(T_mean, u, self.ri)

        # Integrate up the total contributions
        fluid_temps = self.fluid_temperatures(T_start, T_tube, t)
        T_metal = self.metal_temperature(t)

        flux = (
            self.weights[:, None, None]
            * h[:, None, None]
            * (T_metal - fluid_temps[:, None, :])
        )

        return self.ri * self.dz * self.dtheta * jnp.sum(flux, axis=(1, 2))

    def flow_rates(self, T_start, T_tube, t):
        """Recover the flow rates for a given state

        Args:
            T_start:    inlet temperatures
            T_tube:     outlet temperatures
            t:          time
        """
        mdot = self.mass_flow_rate(t)
        T_mean = self.mean_temperature(T_start, T_tube)
        rho = self.material.rho(T_mean)
        return mdot / (self.ntube * np.pi * rho * self.ri**2.0)

    def fluid_temperatures(self, T_start, T_tube, t):
        """Recover the fluid temperatures for a given state

        Args:
            T_start:    inlet temperatures
            T_tube:     outlet temperatures
            t:          time
        """
        return ((T_tube - T_start) / self.h * self.zs[:, None] + T_start).T


def deparameterize_flow_path(pset):
    """
    Convert parameter set to kwargs for FlowPath
    """
    return {
        "rtol": pset.get_default("rtol", 1e-6),
        "atol": pset.get_default("atol", 1e-8),
        "miter": pset.get_default("miter", 100),
        "verbose": pset.get_default("verbose", False),
    }


class FlowPath:
    """Flow path data structure

    A simplified description of a flow path as chain.
    """

    def __init__(
        self,
        times,
        mass_flow,
        inlet_temperature,
        rtol=1e-6,
        atol=1e-8,
        miter=50,
        verbose=False,
        **kwargs
    ):
        """
        Parameters:
            times (ntime,):             discrete times
            mass_flow (ntime,):         mass flow rate at each time
            inlet_temperature (ntime,): inlet temperature at each time
        """
        # Always start with the inlet node!
        self.times = times
        self.mass_flow = mass_flow
        self.inlet_temperature = inlet_temperature
        self.chain = [StartLink(times, inlet_temperature)]

        # Solver parameters
        self.rtol = 1e-6
        self.atol = 1e-8
        self.miter = 100
        self.verbose = verbose

    def add_panel(self, weights, ri, h, metal_temp, material):
        """
        Construct and add the standard panel -> manifold link

        Parameters:
            weights:        tube weights
            ri:             tube inner radius
            h:              tube height
            metal_temp:     metal temperatures
            material:       thermofluid material
        """
        self.chain.append(
            SimplePanelLink(
                self.times, self.mass_flow, weights, ri, h, metal_temp, material
            )
        )
        self.chain.append(ManifoldLink(self.times, self.mass_flow, weights))

    def add_panel_from_object(self, panel, material):
        """Same as add_panel, but use a panel object directly

        Args:
            panel (receiver.Panel): panel object
            material: fluid material properties
        """
        weights = np.array([float(tube.multiplier) for tube in panel.tubes.values()])
        # Fix this, we should take variable radii and heights
        ri = np.array([tube.r - tube.t for tube in panel.tubes.values()])[0]
        h = np.array([tube.h for tube in panel.tubes.values()])[0]

        metal_temps = []
        for tube in panel.tubes.values():
            if tube.abstraction == "3D":
                metal_temps.append(
                    tube.quadrature_results["ghost_temperature"][..., 1, 1:-1, 1:-1]
                )
            elif tube.abstraction == "2D":
                metal_temps.append(
                    np.broadcast_to(
                        tube.quadrature_results["ghost_temperature"][:, 1, 1:-1, None],
                        (tube.ntime, tube.nt, tube.nz),
                    )
                )
            elif tube.abstraction == "1D":
                metal_temps.append(
                    np.broadcast_to(
                        tube.quadrature_results["ghost_temperature"][:, 1, None, None],
                        (tube.ntime, tube.nt, tube.nz),
                    )
                )
            else:
                raise ValueError("Need to add")

        metal_temps = np.swapaxes(np.array(metal_temps), 0, 1)

        self.add_panel(weights, ri, h, metal_temps, material)

    def _setup(self):
        """
        Setup the data structures needed to solve the current chain
        """
        self.nvals = 0
        self.dof_map = []
        for obj in self.chain:
            self.dof_map.append(list(range(self.nvals, self.nvals + obj.size)))
            self.nvals += obj.size

    def solve(self, t):
        """
        Solve for the current fluid temperatures

        Args:
            t:      time
        """
        self._setup()
        self.t = t

        # Significant decision...
        T = np.zeros((self.nvals))

        # Initial residual
        R, J = self.RJ(T)
        nr0 = la.norm(R)

        if self.verbose:
            print("Iter.\t||R||\t\t||R||/||R0||")
            print("%i\t%e" % (0, nr0))

        for i in range(self.miter):
            dT = spla.spsolve(J, R)
            T -= dT
            R, J = self.RJ(T)
            nr = la.norm(R)
            if np.isnan(nr):
                raise RuntimeError("NaN detected!")
            if self.verbose:
                print("%i\t%e\t%e" % (i + 1, nr, nr / nr0))
            if (nr < self.atol) or (nr / nr0 < self.rtol):
                break
        else:
            raise RuntimeError("Too many iterations in newton solver!")

        return T

    def recover_tube_results(self, T, t):
        """Recover the tube flow rates and actual temperature fields

        Args:
            T (np.array):   temperatures
            t (float):      time
        """
        flow_rates = []
        tube_temperatures = []
        dofs_prev = []
        for i, (obj, dofs) in enumerate(zip(self.chain, self.dof_map)):
            # Panels are every other entry
            T_prev = T[dofs_prev]
            T_curr = T[dofs]

            if i % 2 == 1:
                flow_rates.append(obj.flow_rates(T_prev, T_curr, t))
                tube_temperatures.append(obj.fluid_temperatures(T_prev, T_curr, t))
            dofs_prev = dofs

        return flow_rates, tube_temperatures

    def RJ(self, T):
        """Formulate the residual and Jacobian for a given time

        Args:
            T (np.array):   nodal temperatures
        """
        R = np.zeros((self.nvals,))
        I = []
        J = []
        vals = []

        dofs_prev = []
        for obj, dofs in zip(self.chain, self.dof_map):
            T_prev = jnp.asarray(T[dofs_prev])
            T_curr = jnp.asarray(T[dofs])

            R[dofs] = obj.residual(T_prev, T_curr, self.t)
            Jprev = jacfwd(obj.residual, argnums=0)(T_prev, T_curr, self.t)

            i, j, v = coo_entries(dofs, dofs_prev, Jprev)
            I.extend(i)
            J.extend(j)
            vals.extend(v)

            Jcurr = jacfwd(obj.residual, argnums=1)(T_prev, T_curr, self.t)
            i, j, v = coo_entries(dofs, dofs, Jcurr)
            I.extend(i)
            J.extend(j)
            vals.extend(v)

            dofs_prev = dofs

        J = sp.coo_array((vals, (I, J)), shape=(self.nvals, self.nvals))
        return R, J.tocsr()


def coo_entries(row, col, jac):
    """
    Provide flat list for the row indices, column indices, and entries
    of a jacobian entry

    Args:
        row (np.array): row indices
        col (np.array): column indices
        jac (np.array): actual matrix entries
    """
    i, j = np.meshgrid(row, col)

    return list(i.flatten()), list(j.flatten()), list(jac.flatten())



import unittest
import tempfile

import numpy as np
from srlife.thermohydraulics import thermalfluid


class TestPolynomialThermalFluidMaterial(unittest.TestCase):
    def setUp(self):
        self.cp = np.array([0.0000290556, 0.254458])
        self.rho = np.array([-5.52e-10, 2.0544788e-6])
        self.mu = np.array([6.4224e-8, -0.000139846, 0.087281])
        self.k = np.array([-1e-7, 0.000532015])

        self.model = thermalfluid.PolynomialThermalFluidMaterial(
            self.cp, self.rho, self.mu, self.k
        )
        self.T = np.array([100.0, 200.0, 300.0]) + 273.15
        self.u = np.array([1.0, 2.0, 3.0])
        self.r = np.array([5.0, 10.0, 15.0])

    def test_definitions(self):
        self.assertTrue(np.allclose(self.model.cp(self.T), np.polyval(self.cp, self.T)))
        self.assertTrue(
            np.allclose(self.model.rho(self.T), np.polyval(self.rho, self.T))
        )
        self.assertTrue(np.allclose(self.model.mu(self.T), np.polyval(self.mu, self.T)))
        self.assertTrue(np.allclose(self.model.k(self.T), np.polyval(self.k, self.T)))

    def test_recover(self):
        tfile = tempfile.mktemp()
        self.model.save(tfile, "whatever")

        recover = thermalfluid.ThermalFluidMaterial.load(tfile, "whatever")

        self.assertTrue(np.allclose(recover.cp_poly, self.model.cp_poly))
        self.assertTrue(np.allclose(recover.rho_poly, self.model.rho_poly))
        self.assertTrue(np.allclose(recover.mu_poly, self.model.mu_poly))
        self.assertTrue(np.allclose(recover.k_poly, self.model.k_poly))

        self.assertTrue(np.isclose(recover.film_min, self.model.film_min))
        self.assertTrue(np.isclose(recover.T_max, self.model.T_max))
        self.assertTrue(np.isclose(recover.T_min, self.model.T_min))
        self.assertTrue(np.isclose(recover.laminar_cutoff, self.model.laminar_cutoff))
        self.assertTrue(np.isclose(recover.laminar_value, self.model.laminar_value))



import unittest

import numpy as np
import vtk
from vtk.util.numpy_support import vtk_to_numpy

from srlife import receiver, writers


class TestVolumeCalculation(unittest.TestCase):
    def setUp(self):
        self.ro = 9.0
        self.t = 1.2
        self.h = 15.0

        self.nr = 10
        self.nt = 36
        self.nz = 11

        self.tube = receiver.Tube(self.ro, self.t, self.h, self.nr, self.nt, self.nz)

    def test_3D(self):
        v1 = self.tube.element_volumes()

        writer = writers.VTKWriter(self.tube, "dummy.vtk")
        obj = writer.make_vtk_object()

        meshQuality = vtk.vtkMeshQuality()
        meshQuality.SetInputData(obj)
        meshQuality.SetHexQualityMeasureToVolume()
        meshQuality.Update()
        res = meshQuality.GetOutput()
        vols = vtk_to_numpy(res.GetCellData().GetArray("Quality"))

        err = np.abs(v1 - vols) / np.abs(vols)

        self.assertTrue(np.allclose(v1, vols, rtol=1e-4))

    def test_2D(self):
        self.tube.make_2D(self.h / 2)
        v1 = self.tube.element_volumes()

        writer = writers.VTKWriter(self.tube, "dummy.vtk")
        obj = writer.make_vtk_object()

        meshQuality = vtk.vtkMeshQuality()
        meshQuality.SetInputData(obj)
        meshQuality.SetQuadQualityMeasureToArea()
        meshQuality.Update()
        res = meshQuality.GetOutput()
        vols = vtk_to_numpy(res.GetCellData().GetArray("Quality")) * self.h

        self.assertTrue(np.allclose(v1, vols, rtol=1e-4))

    def test_1D(self):
        self.tube.make_1D(self.h / 2, 0)
        v1 = self.tube.element_volumes()
        vt = np.sum(v1)

        v_check = np.pi * (self.ro**2.0 - (self.ro - self.t) ** 2.0) * self.h

        self.assertAlmostEqual(vt, v_check)



import unittest

import os.path

# Shut up matplotlib
import logging

mpl_logger = logging.getLogger("matplotlib")
mpl_logger.setLevel(logging.WARNING)

import numpy as np

import sys

sys.path.append("test/solvers/tube")
sys.path.append("../../..")
from srlife import structural, receiver

from neml import elasticity, models, parse

from run_structural_tube_verification import cases, do_complete_comparison


def differentiate(fn, x0, eps=1.5e-6):
    if np.isscalar(x0):
        sx = (1,)
    else:
        sx = x0.shape

    f0 = fn(x0)

    if np.isscalar(f0):
        sf = (1,)
    else:
        sf = f0.shape

    tshape = sf + sx
    rshape = sx + sf

    D = np.zeros(rshape)

    for index in np.ndindex(sx):
        if np.isscalar(x0):
            diff = np.abs(x0 * eps)
        else:
            diff = np.abs(x0[index] * eps)
        if diff < eps:
            diff = eps

        if np.isscalar(x0):
            xp = diff
        else:
            xp = np.zeros(sx)
            xp[index] = diff

        fp = fn(xp + x0)

        D[index] = (fp - f0) / diff

    Df = np.zeros(tshape)
    # Reverse
    for ind1 in np.ndindex(sx):
        for ind2 in np.ndindex(sf):
            Df[ind2 + ind1] = D[ind1 + ind2]

    if np.isscalar(x0):
        return Df[0][0]
    else:
        return Df


class TestSimpleCases(unittest.TestCase):
    def setUp(self):
        self.atol = 1.0e-3
        self.rtol = 1.0e-3

        self.problems = cases
        self.solver = structural.PythonTubeSolver()

    def _check_case(self, d, case):
        tube = case.run_comparison(d, self.solver)
        a, r = case.evaluate_comparison(tube)

        self.assertTrue(a < self.atol or r < self.rtol)

    def test_all(self):
        for d in range(1, 4):
            for c in self.problems:
                self._check_case(d, c)


class TestMOOSEComparison(unittest.TestCase):
    def setUp(self):
        self.solver = structural.PythonTubeSolver()
        self.atol = 1.0e-3

    def test_1D(self):
        a, r = do_complete_comparison(1, self.solver)
        self.assertTrue(a < self.atol)

    def test_2D(self):
        a, r = do_complete_comparison(2, self.solver)
        self.assertTrue(a < self.atol)

    # Too heavy right now
    # def test_3D(self):
    #  a, r = do_complete_comparison(3, self.solver)
    #  self.assertTrue(a < self.atol)


class TestAxialStiffnessExact(unittest.TestCase):
    def setUp(self):
        self.ro = 5
        self.ri = 4.5
        self.h = 2.5

        self.nr = 10
        self.nt = 20
        self.nz = 5

        self.E = 150000.0
        self.nu = 0.35

        self.tube = receiver.Tube(
            self.ro, self.ro - self.ri, self.h, self.nr, self.nt, self.nz
        )
        self.times = np.array([0, 1])
        self.tube.set_times(self.times)

        self.tube.set_pressure_bc(receiver.PressureBC(self.times, self.times * 0))

        self.emodel = elasticity.IsotropicLinearElasticModel(
            self.E, "youngs", self.nu, "poissons"
        )
        self.mat = models.SmallStrainElasticity(self.emodel)

        self.d = 0.25

        self.force_exact = (
            np.pi * (self.ro**2.0 - self.ri**2.0) * self.E * self.d / self.h
        )
        self.stiffness_exact = (
            np.pi * (self.ro**2.0 - self.ri**2.0) * self.E / self.h
        )

        self.solver = structural.PythonTubeSolver(verbose=False)

    def _solve(self, d):
        self.solver.setup_tube(self.tube)
        state_n = self.solver.init_state(self.tube, self.mat)

        return self.solver.solve(self.tube, 1, state_n, d)

    def test_1D(self):
        self.tube.make_1D(self.h / 2, 0)

        state = self._solve(self.d)

        self.assertAlmostEqual(self.force_exact, state.force, places=3)
        self.assertAlmostEqual(self.stiffness_exact, state.stiffness, places=3)

    def test_2D(self):
        self.tube.make_2D(self.h / 2)

        state = self._solve(self.d)

        cf = np.sum(state.basis.dx) / (np.pi * (self.ro**2.0 - self.ri**2.0))

        self.assertAlmostEqual(self.force_exact, state.force / cf, places=3)
        self.assertAlmostEqual(self.stiffness_exact, state.stiffness / cf, places=3)

    def test_3D(self):
        state = self._solve(self.d)

        cf = (
            np.sum(state.basis.dx)
            / self.h
            / (np.pi * (self.ro**2.0 - self.ri**2.0))
        )

        self.assertAlmostEqual(self.force_exact, state.force / cf, places=3)
        self.assertAlmostEqual(self.stiffness_exact, state.stiffness / cf, places=3)


class TestAxialStiffnessNumerical(unittest.TestCase):
    def setUp(self):
        self.ro = 5
        self.ri = 4.5
        self.h = 2.5

        self.nr = 10
        self.nt = 20
        self.nz = 5

        self.E = 150000.0
        self.nu = 0.35

        self.tube = receiver.Tube(
            self.ro, self.ro - self.ri, self.h, self.nr, self.nt, self.nz
        )
        self.times = np.array([0, 1])
        self.tube.set_times(self.times)

        self.tube.set_pressure_bc(receiver.PressureBC(self.times, self.times / 50.0))

        self.mat = parse.parse_xml(
            os.path.join(os.path.dirname(__file__), "moose-verification", "model.xml"),
            "creeping",
        )

        self.d = 0.25

        self.solver = structural.PythonTubeSolver(verbose=False)

    def _solve(self, d):
        self.solver.setup_tube(self.tube)
        state_n = self.solver.init_state(self.tube, self.mat)

        return self.solver.solve(self.tube, 1, state_n, d)

    def test_1D(self):
        self.tube.make_1D(self.h / 2, 0)

        exact = self._solve(self.d).stiffness
        numerical = differentiate(lambda d: self._solve(d).force, self.d)

        self.assertTrue(np.isclose(exact, numerical, rtol=1e-4))

    def test_2D(self):
        self.tube.make_2D(self.h / 2)

        exact = self._solve(self.d).stiffness
        numerical = differentiate(lambda d: self._solve(d).force, self.d)

        self.assertTrue(np.isclose(exact, numerical, rtol=1e-4))

    # Too heavy
    def test_3D(self):
        exact = self._solve(self.d).stiffness
        numerical = differentiate(lambda d: self._solve(d).force, self.d)

        self.assertTrue(np.isclose(exact, numerical, rtol=1e-4))
